# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from collections import deque
from unittest.mock import call, patch

import pytest

from oumi.inference.adaptive_semaphore import AdaptiveSemaphore, PoliteAdaptiveSemaphore


@pytest.fixture
def mock_time():
    with patch("oumi.inference.adaptive_semaphore.time") as time_mock:
        yield time_mock


@pytest.fixture
def mock_asyncio_sleep():
    with patch("oumi.inference.adaptive_semaphore.asyncio.sleep") as sleep_mock:
        yield sleep_mock


@pytest.mark.asyncio
async def test_polite_adaptive_semaphore(mock_time, mock_asyncio_sleep):
    semaphore = PoliteAdaptiveSemaphore(capacity=1, politeness_policy=10)
    mock_time.time.return_value = 1
    await semaphore.acquire()
    assert len(semaphore._queue) == 0
    semaphore.release()
    # 11 = 1.0 + politeness_policy  (10.0)
    assert semaphore._queue == deque([11])


@pytest.mark.asyncio
async def test_polite_adaptive_semaphore_adjust_capacity(mock_time, mock_asyncio_sleep):
    semaphore = PoliteAdaptiveSemaphore(capacity=1, politeness_policy=10)
    mock_time.time.return_value = 0.0
    # Initially the semaphore has capacity of 1 and is unused, so the queue is [-1].
    assert semaphore._queue == deque([-1])
    await semaphore.adjust_capacity(2)
    # We expanded the capacity to 2, so the queue is padded with -1s.
    assert semaphore._queue == deque([-1, -1])
    mock_time.time.return_value = 1.0
    await semaphore.acquire()
    semaphore.release()
    mock_time.time.return_value = 2.0
    # We acquired the first permit (consumed -1) then released (appended the time +
    # politeness policy=10.0). The queue is now [-1, 11].
    assert semaphore._queue == deque([-1, 11])
    await semaphore.acquire()
    semaphore.release()
    mock_time.time.return_value = 3.0
    # We acquired the second permit (consumed -1) then released (appended the time +
    # politeness policy=10.0). The queue is now [11, 12].
    assert semaphore._queue == deque([11, 12])
    await semaphore.adjust_capacity(1)
    # We reduced the capacity to 1, so the queue is now [12].
    # 11 was removed from the queue because it was the oldest entry.
    assert semaphore._queue == deque([12])
    await semaphore.acquire()
    # We acquired the third permit (consumed 12). The queue is now empty.
    assert semaphore._queue == deque([])
    await semaphore.adjust_capacity(10)
    # We expanded the capacity to 10, so the queue is padded with -1s.
    assert semaphore._queue == deque([-1] * 10)


@pytest.mark.asyncio
async def test_polite_adaptive_semaphore_subsequent_acquires_use_queue(
    mock_time, mock_asyncio_sleep
):
    """Test that subsequent acquires use the queue values for waiting."""
    semaphore = PoliteAdaptiveSemaphore(capacity=2, politeness_policy=2.0)

    # First acquire - no wait needed since queue is empty
    mock_time.time.return_value = 10.0
    await semaphore.acquire()
    assert semaphore._queue == deque([-1])

    # Release - adds current time to queue
    semaphore.release()
    assert semaphore._queue == deque([-1, 12.0])

    # Second acquire - should wait based on queue value
    mock_time.time.return_value = 11.0  # 1 second later
    await semaphore.acquire()

    assert semaphore._queue == deque([12])


@pytest.mark.asyncio
async def test_polite_adaptive_semaphore_multiple_releases_before_acquire(
    mock_time, mock_asyncio_sleep
):
    """Test that multiple releases create a queue that subsequent acquires respect."""
    semaphore = PoliteAdaptiveSemaphore(capacity=2, politeness_policy=20)

    mock_time.time.side_effect = [
        10.0,
        11.0,
        12.0,
        36.0,  # release time for Task 3
        37.0,
        38.0,
    ]

    async def acquire_and_release(semaphore: PoliteAdaptiveSemaphore):
        await semaphore.acquire()
        semaphore.release()

    tasks = [acquire_and_release(semaphore) for _ in range(4)]

    # Task 1: Acquires at first available slot and finishes at 10.0. Sets slot to next
    # allowed start at 10.0 + 20 = 30.0.
    # Task 2: Acquires the other slot, finishes at 11.0. Sets slot to next allowed start
    # at 11.0 + 20 = 31.0
    # Task 3: Starts at 12.0. Both queue slots are now [30.0, 31.0]. Needs to wait for
    # queue.pop() - current time = 30.0 - 12.0 = 18.0 seconds.
    # Task 4: Starts at 37.0. The queue slots are now [31.0, 56.0]
    # (after Task 3’s release). 37.0 is after 31.0, so no sleep needed. Task 4 releases
    # at 38.0, setting queue to [56.0, 58.0].

    await asyncio.gather(*tasks)
    mock_time.time.assert_has_calls([call(), call(), call(), call(), call(), call()])
    mock_asyncio_sleep.assert_has_calls([call(18.0)])
    assert semaphore._queue == deque([56.0, 58.0])


@pytest.mark.asyncio
async def test_polite_adaptive_semaphore_increments_queue(
    mock_time, mock_asyncio_sleep
):
    """Test that queue is updated with each release and acquire."""
    semaphore = PoliteAdaptiveSemaphore(capacity=2, politeness_policy=1.0)

    # First acquire and release
    mock_time.time.return_value = 10.0
    await semaphore.acquire()
    semaphore.release()
    assert semaphore._queue == deque([-1, 11.0])

    # Second acquire and release
    mock_time.time.return_value = 11.0
    await semaphore.acquire()
    semaphore.release()
    assert semaphore._queue == deque([11.0, 12.0])

    mock_time.time.return_value = 12.0
    await semaphore.acquire()
    semaphore.release()
    assert semaphore._queue == deque([12.0, 13.0])

    mock_time.time.return_value = 13.0
    await semaphore.acquire()

    mock_asyncio_sleep.assert_not_called()


@pytest.mark.asyncio
async def test_polite_adaptive_semaphore_wait_time_calculation(
    mock_time, mock_asyncio_sleep
):
    """Test specific wait time calculations."""
    semaphore = PoliteAdaptiveSemaphore(capacity=1, politeness_policy=3.0)

    # First acquire and release
    mock_time.time.return_value = 100.0
    await semaphore.acquire()
    semaphore.release()
    assert semaphore._queue == deque([103.0])

    # Second acquire - should wait exactly 3 seconds
    mock_time.time.return_value = 101.0  # 1 second later
    await semaphore.acquire()

    # Should wait: (100.0 + 3.0) - 101.0 = 2.0 seconds
    mock_asyncio_sleep.assert_called_once_with(2.0)


@pytest.mark.asyncio
async def test_polite_adaptive_semaphore_no_wait_when_enough_time_passed(
    mock_time, mock_asyncio_sleep
):
    """Test that no wait occurs when enough time has passed since last release."""
    semaphore = PoliteAdaptiveSemaphore(capacity=1, politeness_policy=2.0)

    # First acquire and release
    mock_time.time.return_value = 10.0
    await semaphore.acquire()
    semaphore.release()
    assert semaphore._queue == deque([12.0])

    # Second acquire - wait more than politeness policy
    mock_time.time.return_value = 13.0  # 3 seconds later (more than 2.0 politeness)
    await semaphore.acquire()

    # Should not wait since enough time has passed
    mock_asyncio_sleep.assert_not_called()


@pytest.mark.asyncio
async def test_polite_adaptive_semaphore_concurrent_acquires(
    mock_time, mock_asyncio_sleep
):
    """Test that concurrent acquires respect the politeness policy."""
    semaphore = PoliteAdaptiveSemaphore(capacity=2, politeness_policy=1.0)

    # First two acquires should happen immediately
    mock_time.time.return_value = 10.0
    await semaphore.acquire()
    await semaphore.acquire()

    # Release first one
    semaphore.release()
    assert semaphore._queue == deque([11.0])

    # Release second one
    mock_time.time.return_value = 11.0
    semaphore.release()
    assert semaphore._queue == deque([11.0, 12.0])

    # Next acquire should wait based on oldest queue value
    mock_time.time.return_value = 12.0
    await semaphore.acquire()

    # Should wait: (10.0 + 1.0) - 12.0 = -1.0 (no wait needed)
    mock_asyncio_sleep.assert_not_called()


@pytest.mark.asyncio
async def test_polite_adaptive_semaphore_empty_queue_no_wait(
    mock_time, mock_asyncio_sleep
):
    """Test that acquire with empty queue doesn't wait."""
    semaphore = PoliteAdaptiveSemaphore(capacity=1, politeness_policy=5.0)

    # First acquire - no wait needed
    mock_time.time.return_value = 100.0
    await semaphore.acquire()

    # Should not sleep since queue is empty
    mock_asyncio_sleep.assert_not_called()


class TestAdaptiveSemaphore:
    """Test suite for AdaptiveSemaphore class."""

    @pytest.mark.asyncio
    async def test_initialization(self):
        """Test semaphore initialization with different capacities."""
        # Test basic initialization
        semaphore = AdaptiveSemaphore(5)
        assert semaphore._max_capacity == 5
        assert semaphore._current_capacity == 5
        assert len(semaphore._waiters) == 0

        # Test with capacity of 1
        semaphore = AdaptiveSemaphore(1)
        assert semaphore._max_capacity == 1
        assert semaphore._current_capacity == 1

    @pytest.mark.asyncio
    async def test_basic_acquire_release(self):
        """Test basic acquire and release functionality."""
        semaphore = AdaptiveSemaphore(2)

        # First acquire should succeed immediately
        await semaphore.acquire()
        assert semaphore._current_capacity == 1

        # Second acquire should succeed immediately
        await semaphore.acquire()
        assert semaphore._current_capacity == 0

        # Release one permit
        semaphore.release()
        assert semaphore._current_capacity == 1

        # Release another permit
        semaphore.release()
        assert semaphore._current_capacity == 2

    @pytest.mark.asyncio
    async def test_acquire_blocking_when_capacity_reached(self):
        """Test that acquire blocks when capacity is reached."""
        semaphore = AdaptiveSemaphore(1)

        # First acquire should succeed immediately
        await semaphore.acquire()
        assert semaphore._current_capacity == 0

        # Second acquire should block
        acquire_task = asyncio.create_task(semaphore.acquire())
        await asyncio.sleep(0.01)  # Give task a chance to run
        assert not acquire_task.done()
        assert len(semaphore._waiters) == 1

        # Release should unblock the waiting acquire
        semaphore.release()
        await acquire_task  # This should complete now
        assert semaphore._current_capacity == 0

    @pytest.mark.asyncio
    async def test_multiple_waiters(self):
        """Test handling of multiple waiting tasks."""
        semaphore = AdaptiveSemaphore(1)

        # Acquire the only permit
        await semaphore.acquire()
        assert semaphore._current_capacity == 0

        # Create multiple waiting tasks
        task1 = asyncio.create_task(semaphore.acquire())
        task2 = asyncio.create_task(semaphore.acquire())
        task3 = asyncio.create_task(semaphore.acquire())

        await asyncio.sleep(0.01)  # Let tasks start waiting
        assert len(semaphore._waiters) == 3
        assert not task1.done()
        assert not task2.done()
        assert not task3.done()

        # Release should wake up first waiter
        semaphore.release()
        await task1  # Should complete
        assert not task2.done()
        assert not task3.done()
        assert len(semaphore._waiters) == 2

        # Release again should wake up second waiter
        semaphore.release()
        await task2  # Should complete
        assert not task3.done()
        assert len(semaphore._waiters) == 1

        # Release again should wake up third waiter
        semaphore.release()
        await task3  # Should complete
        assert len(semaphore._waiters) == 0

    @pytest.mark.asyncio
    async def test_cancelled_waiters_handling(self):
        """Test that cancelled waiters are properly removed."""
        semaphore = AdaptiveSemaphore(1)

        # Acquire the only permit
        await semaphore.acquire()

        # Create waiting tasks
        task1 = asyncio.create_task(semaphore.acquire())
        task2 = asyncio.create_task(semaphore.acquire())

        await asyncio.sleep(0.01)
        assert len(semaphore._waiters) == 2

        # Cancel first task
        task1.cancel()
        try:
            await task1
        except asyncio.CancelledError:
            pass

        # Release should wake up the non-cancelled waiter
        semaphore.release()
        await task2  # Should complete
        assert len(semaphore._waiters) == 0

    @pytest.mark.asyncio
    async def test_adjust_capacity_increase(self):
        """Test increasing semaphore capacity."""
        semaphore = AdaptiveSemaphore(2)

        # Acquire both permits
        await semaphore.acquire()
        await semaphore.acquire()
        assert semaphore._current_capacity == 0

        # Create waiting tasks
        task1 = asyncio.create_task(semaphore.acquire())
        task2 = asyncio.create_task(semaphore.acquire())

        await asyncio.sleep(0.01)
        assert len(semaphore._waiters) == 2

        # Increase capacity should wake up waiters
        await semaphore.adjust_capacity(4)
        assert semaphore._max_capacity == 4
        await asyncio.sleep(0.01)

        await task1  # Should complete
        await task2  # Should complete
        assert len(semaphore._waiters) == 0
        assert semaphore._current_capacity == 0  # 4 capacity - 4 acquired

    @pytest.mark.asyncio
    async def test_adjust_capacity_decrease(self):
        """Test decreasing semaphore capacity."""
        semaphore = AdaptiveSemaphore(5)

        # Acquire 2 permits
        await semaphore.acquire()
        await semaphore.acquire()
        assert semaphore._current_capacity == 3

        # Decrease capacity
        await semaphore.adjust_capacity(3)
        assert semaphore._max_capacity == 3
        assert semaphore._current_capacity == 1  # 3 - 2 acquired

        # Should be able to acquire one more
        await semaphore.acquire()
        assert semaphore._current_capacity == 0

        # Next acquire should block
        task = asyncio.create_task(semaphore.acquire())
        await asyncio.sleep(0.01)
        assert not task.done()
        assert len(semaphore._waiters) == 1

        # Clean up
        semaphore.release()
        await task

    @pytest.mark.asyncio
    async def test_adjust_capacity_with_existing_waiters(self):
        """Test capacity adjustment when there are existing waiters."""
        semaphore = AdaptiveSemaphore(1)

        # Acquire the permit
        await semaphore.acquire()

        # Create multiple waiters
        tasks = [asyncio.create_task(semaphore.acquire()) for _ in range(3)]
        await asyncio.sleep(0.01)
        assert len(semaphore._waiters) == 3

        # Increase capacity to 3 should wake up 2 waiters
        await semaphore.adjust_capacity(3)
        await asyncio.sleep(0.01)

        # First two tasks should complete, third should still wait
        await tasks[0]
        await tasks[1]
        assert not tasks[2].done()
        assert len(semaphore._waiters) == 1

        # Release one more to complete the last task
        semaphore.release()
        await tasks[2]

    @pytest.mark.asyncio
    async def test_adjust_capacity_decrease_below_current_usage(self):
        """Test decreasing capacity below current usage doesn't revoke permits."""
        semaphore = AdaptiveSemaphore(5)

        # Acquire 4 permits
        for _ in range(4):
            await semaphore.acquire()
        assert semaphore._current_capacity == 1

        # Decrease capacity below current usage
        await semaphore.adjust_capacity(2)
        assert semaphore._max_capacity == 2
        # Current count should never go below 0
        assert semaphore._current_capacity == 0

        # No new acquires should succeed until enough permits are released
        task = asyncio.create_task(semaphore.acquire())
        await asyncio.sleep(0.01)
        assert not task.done()

        # Release 3 permits to get back to positive count
        for _ in range(3):
            semaphore.release()

        # Now the waiting task should complete
        await task

    @pytest.mark.asyncio
    async def test_concurrent_operations(self):
        """Test concurrent acquire/release/adjust operations."""
        semaphore = AdaptiveSemaphore(3)

        async def worker(worker_id: int, results: list):
            try:
                await semaphore.acquire()
                await asyncio.sleep(0.01)  # Simulate work
                results.append(f"worker_{worker_id}_done")
                semaphore.release()
            except Exception as e:
                results.append(f"worker_{worker_id}_error_{e}")

        # Start multiple workers
        results = []
        tasks = [asyncio.create_task(worker(i, results)) for i in range(5)]

        # Adjust capacity while workers are running
        await asyncio.sleep(0.005)
        await semaphore.adjust_capacity(2)
        await asyncio.sleep(0.005)
        await semaphore.adjust_capacity(4)

        # Wait for all workers to complete
        await asyncio.gather(*tasks)

        # All workers should complete successfully
        assert len(results) == 5
        assert all("_done" in result for result in results)

    @pytest.mark.asyncio
    async def test_zero_capacity(self):
        """Test semaphore behavior with zero capacity."""
        semaphore = AdaptiveSemaphore(1)

        # Reduce to zero capacity
        with pytest.raises(ValueError):
            await semaphore.adjust_capacity(0)

        with pytest.raises(ValueError):
            await semaphore.adjust_capacity(-1)

        with pytest.raises(ValueError):
            semaphore = AdaptiveSemaphore(0)

    @pytest.mark.asyncio
    async def test_edge_case_empty_waiters_on_release(self):
        """Test release when there are no waiters."""
        semaphore = AdaptiveSemaphore(2)

        # Acquire one permit
        await semaphore.acquire()
        assert semaphore._current_capacity == 1

        # Release without any waiters
        semaphore.release()
        assert semaphore._current_capacity == 2
        assert len(semaphore._waiters) == 0

    @pytest.mark.asyncio
    async def test_waiter_cancellation_during_release(self):
        """Test handling of waiter cancellation during release."""
        semaphore = AdaptiveSemaphore(1)

        # Acquire the permit
        await semaphore.acquire()

        # Create waiters
        task1 = asyncio.create_task(semaphore.acquire())
        task2 = asyncio.create_task(semaphore.acquire())

        await asyncio.sleep(0.01)
        assert len(semaphore._waiters) == 2

        # Cancel first waiter
        task1.cancel()

        # Release should skip cancelled waiter and wake up second one
        semaphore.release()

        try:
            await task1
        except asyncio.CancelledError:
            pass

        await task2  # Should complete
        assert len(semaphore._waiters) == 0

    @pytest.mark.asyncio
    async def test_context_manager_basic(self):
        """Test basic context manager functionality."""
        semaphore = AdaptiveSemaphore(2)

        async with semaphore:
            assert semaphore._current_capacity == 1

        assert semaphore._current_capacity == 2

    @pytest.mark.asyncio
    async def test_context_manager_exception_handling(self):
        """Test context manager releases permit even when exception occurs."""
        semaphore = AdaptiveSemaphore(2)

        try:
            async with semaphore:
                assert semaphore._current_capacity == 1
                raise ValueError("Test exception")
        except ValueError:
            pass

        assert semaphore._current_capacity == 2

    @pytest.mark.asyncio
    async def test_over_release_behavior(self):
        """Test behavior when releasing more permits than acquired."""
        semaphore = AdaptiveSemaphore(2)

        # Release without acquiring
        semaphore.release()

        # Current capacity should not exceed max capacity, no error should be raised
        assert semaphore._current_capacity == 2

    def test_string_representation(self):
        """Test string representation of semaphore."""
        semaphore = AdaptiveSemaphore(5)
        repr_str = repr(semaphore)
        assert "AdaptiveSemaphore" in repr_str
        assert "capacity=5" in repr_str
        assert "current_count=5" in repr_str

    @pytest.mark.asyncio
    async def test_locked_method(self):
        """Test locked() method returns correct state."""
        semaphore = AdaptiveSemaphore(2)

        assert not semaphore.locked()

        await semaphore.acquire()
        assert not semaphore.locked()

        await semaphore.acquire()
        assert semaphore.locked()

        semaphore.release()
        assert not semaphore.locked()

    @pytest.mark.asyncio
    async def test_concurrent_capacity_adjustments(self):
        """Test concurrent capacity adjustments don't cause race conditions."""
        semaphore = AdaptiveSemaphore(2)

        async def adjust_up():
            await semaphore.adjust_capacity(5)

        async def adjust_down():
            await semaphore.adjust_capacity(1)

        # Run concurrent adjustments
        await asyncio.gather(adjust_up(), adjust_down())

        # Final state should be consistent
        assert semaphore._max_capacity in [1, 5]
        assert semaphore._current_capacity <= semaphore._max_capacity

    @pytest.mark.asyncio
    async def test_many_waiters_performance(self):
        """Test handling of large numbers of waiters."""
        semaphore = AdaptiveSemaphore(1)

        # Acquire the permit
        await semaphore.acquire()

        # Create many waiters
        num_waiters = 1000
        tasks = [asyncio.create_task(semaphore.acquire()) for _ in range(num_waiters)]

        await asyncio.sleep(0.01)
        assert len(semaphore._waiters) == num_waiters

        # Increase capacity significantly to wake them all
        await semaphore.adjust_capacity(num_waiters + 1)

        # All should complete
        await asyncio.gather(*tasks)
        assert len(semaphore._waiters) == 0

    @pytest.mark.asyncio
    async def test_waiter_cleanup_on_exception(self):
        """Test that waiters are properly cleaned up when tasks fail."""
        semaphore = AdaptiveSemaphore(1)

        await semaphore.acquire()

        async def failing_acquire():
            await semaphore.acquire()
            raise RuntimeError("Task failed after acquire")

        task = asyncio.create_task(failing_acquire())

        # Release to let the task acquire
        semaphore.release()

        with pytest.raises(RuntimeError):
            await task

        # Semaphore should be in clean state
        assert len(semaphore._waiters) == 0
        assert semaphore._current_capacity == 0  # Still acquired by failed task



from importlib.util import find_spec
from unittest.mock import patch

import pytest

from oumi.core.configs import (
    GenerationParams,
    InferenceConfig,
    ModelParams,
    RemoteParams,
)
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.inference.bedrock_inference_engine import BedrockInferenceEngine

boto3_import_failed = find_spec("boto3") is None


@pytest.fixture
def mock_boto3():
    """Mock boto3 client to avoid actual AWS calls."""
    with patch("oumi.inference.bedrock_inference_engine.boto3") as mock:
        yield mock


@pytest.fixture
def bedrock_engine(mock_boto3):
    """Create BedrockInferenceEngine with mocked boto3."""
    model_params = ModelParams(model_name="claude-3")
    remote_params = RemoteParams(api_key="test_api_key")
    return BedrockInferenceEngine(
        model_params=model_params,
        remote_params=remote_params,
    )


@pytest.mark.skipif(boto3_import_failed, reason="boto3 not available")
def test_initialization(mock_boto3):
    model_params = ModelParams(
        model_name="claude-3",
        model_max_length=2048,
    )
    remote_params = RemoteParams(api_key="test-key")
    engine = BedrockInferenceEngine(
        model_params=model_params,
        remote_params=remote_params,
    )

    assert engine._model_params.model_name == "claude-3"
    assert engine._remote_params.api_key == "test-key"


@pytest.mark.skipif(boto3_import_failed, reason="boto3 not available")
def test_convert_conversation_to_api_input(bedrock_engine):
    conversation = Conversation(
        messages=[
            Message(content="System message", role=Role.SYSTEM),
            Message(content="User message", role=Role.USER),
            Message(content="Assistant message", role=Role.ASSISTANT),
        ]
    )
    generation_params = GenerationParams(max_new_tokens=100)

    result = bedrock_engine._convert_conversation_to_api_input(
        conversation, generation_params, bedrock_engine._model_params
    )

    print(result)

    assert result["inferenceConfig"]["maxTokens"] == 100
    assert result["inferenceConfig"]["temperature"] == 0.0
    assert result["inferenceConfig"]["topP"] == 1.0
    assert result["messages"][0]["content"][0]["text"] == "User message"
    assert result["messages"][0]["role"] == "user"
    assert result["messages"][1]["content"][0]["text"] == "Assistant message"
    assert result["messages"][1]["role"] == "assistant"
    assert result["system"][0]["text"] == "System message"


@pytest.mark.skipif(boto3_import_failed, reason="boto3 not available")
def test_convert_api_output_to_conversation(bedrock_engine):
    original_conversation = Conversation(
        messages=[
            Message(content="User message", role=Role.USER),
        ],
        metadata={"key": "value"},
        conversation_id="test_id",
    )
    api_response = {
        "output": {"message": {"content": [{"text": "Assistant response"}]}}
    }

    result = bedrock_engine._convert_api_output_to_conversation(
        api_response, original_conversation
    )

    assert len(result.messages) == 2
    assert result.messages[0].content == "User message"
    assert result.messages[1].content == "Assistant response"
    assert result.messages[1].role == Role.ASSISTANT
    assert result.metadata == {"key": "value"}
    assert result.conversation_id == "test_id"


@pytest.mark.skipif(boto3_import_failed, reason="boto3 not available")
def test_infer_online(bedrock_engine):
    with patch.object(bedrock_engine, "_infer") as mock_infer:
        mock_infer.return_value = [
            Conversation(
                conversation_id="1",
                messages=[Message(content="Response", role=Role.ASSISTANT)],
            )
        ]

        input_conversations = [
            Conversation(
                conversation_id="1",
                messages=[Message(content="Hello", role=Role.USER)],
            )
        ]
        inference_config = InferenceConfig(
            generation=GenerationParams(max_new_tokens=50),
        )

        result = bedrock_engine.infer(input_conversations, inference_config)

        mock_infer.assert_called_once_with(input_conversations, inference_config)
        assert result == mock_infer.return_value



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from enum import Enum
from typing import Literal, Optional

from peft import LoraConfig
from peft.utils.peft_types import TaskType
from transformers import BitsAndBytesConfig

from oumi.core.configs.params.base_params import BaseParams


class PeftSaveMode(Enum):
    """Enum representing how to save the final model during PEFT training.

    While models saved with any of these options can be loaded by Oumi, those saved
    with `ADAPTER_ONLY` are not self-contained; the base model will be loaded
    separately from the local HF cache or downloaded from HF Hub if not in the cache.
    """

    ADAPTER_ONLY = "adapter_only"
    """Only save the model adapter.

    Note that when loading this saved model, the base model will be loaded separately
    from the local HF cache or downloaded from HF Hub.
    """

    ADAPTER_AND_BASE_MODEL = "adapter_and_base_model"
    """Save the base model in addition to the adapter.

    This is similar to `ADAPTER_ONLY`, but the base model's weights are also saved in
    the same directory as the adapter weights, making the output dir self-contained.
    """

    MERGED = "merged"
    """Merge the adapter and base model's weights and save as a single model.

    Note that the resulting model is a standard HF Transformers model, and is no longer
    a PEFT model. A copy of the adapter before merging is saved in the "adapter/"
    subdirectory.
    """


class LoraWeightInitialization(str, Enum):
    """Enum representing the supported weight initializations for LoRA adapters."""

    DEFAULT = "default"  # Use the model reference initialization from Microsoft.
    RANDOM = "random"  # Fully random initialization, discouraged.
    GAUSSIAN = "gaussian"
    EVA = "eva"
    PISA = "pissa"
    PISSA_NITER = "pissa_niter_[number of iters]"
    LOFTQ = "loftq"
    OLORA = "olora"

    def get_literal_value(
        self,
    ) -> Literal[
        "default",
        "random",
        "gaussian",
        "eva",
        "pissa",
        "pissa_niter_[number of iters]",
        "loftq",
        "olora",
    ]:
        """Returns a literal value of the enum."""
        if self.value not in {
            "default",
            "random",
            "gaussian",
            "eva",
            "pissa",
            "pissa_niter_[number of iters]",
            "loftq",
            "olora",
        }:
            raise ValueError(f"Invalid enum value: {self.value}")

        return self.value


@dataclass
class PeftParams(BaseParams):
    # Lora Params
    lora_r: int = field(
        default=8,
        metadata={"help": "LoRA R value."},
    )
    """The rank of the update matrices in LoRA.

    A higher value allows for more expressive adaptations but increases
    the number of trainable parameters.
    """

    lora_alpha: int = field(
        default=8,
        metadata={"help": "LoRA alpha."},
    )
    """The scaling factor for the LoRA update.

    This value is typically set equal to `lora_r` or `2*lora_r` for stable training.
    """

    lora_dropout: float = field(
        default=0.0,
        metadata={"help": "LoRA dropout."},
    )
    """The dropout probability applied to LoRA layers.

    This helps prevent overfitting in the adaptation layers.
    """

    lora_target_modules: Optional[list[str]] = field(
        default=None,
        metadata={"help": "LoRA target modules."},
    )
    """List of module names/regexes to apply LoRA to.

    If None, modules that are LoRA-trained are chosen based on the model's architecture.
    Specify ["all-linear"] to apply LoRA to all linear/Conv1D layers in the model.
    Specify a list of module names to only apply LoRA to those modules in the model.
    Finally, specifying [] to avoid targeting any modules (ex. if you want to set
    lora_target_parameters instead).
    """

    lora_target_parameters: Optional[list[str]] = field(
        default=None,
        metadata={"help": "LoRA target parameters."},
    )
    """List of parameter names/regexes to apply LoRA to.

    This is similar to `lora_target_modules` (which you should prefer using if possible)
    but for parameters instead of modules. This is generally only useful for models like
    MoEs that sometimes use nn.Parameter instead of nn.Module.
    """

    lora_modules_to_save: Optional[list[str]] = field(
        default=None,
        metadata={"help": "Model layers to unfreeze and train."},
    )
    """List of module names to unfreeze and train alongside LoRA parameters.

    These modules will be fully fine-tuned, not adapted using LoRA.
    Use this to selectively train certain parts of the model in full precision.
    """

    lora_bias: str = field(
        default="none",
        metadata={
            "help": (
                "Bias type for Lora. Can be 'none', 'all' or 'lora_only'. "
                "If 'all' or 'lora_only', the corresponding biases will "
                "be updated during training. Be aware that this means that, "
                "even when disabling the adapters, the model will not "
                "produce the same output as the base model would have "
                "without adaptation."
                "NOTE: see: "
                "https://github.com/huggingface/peft/blob/main/src/peft/tuners/lora/config.py"
                "for more details."
            )
        },
    )
    """Bias type for LoRA.

    Can be 'none', 'all' or 'lora_only':
    - 'none': No biases are trained.
    - 'all': All biases in the model are trained.
    - 'lora_only': Only biases in LoRA layers are trained.

    If 'all' or 'lora_only', the corresponding biases will be updated during training.
    Note that this means even when disabling the adapters, the model will not produce
    the same output as the base model would have without adaptation.

    For more details, see:
    https://github.com/huggingface/peft/blob/main/src/peft/tuners/lora/config.py
    """

    lora_init_weights: LoraWeightInitialization = field(
        default=LoraWeightInitialization.DEFAULT,
        metadata={
            "help": "Weights initialization for LoRA adapters.",
        },
    )
    """
    Passing `LoraWeightInitialization.DEFAULT` will use the underlying reference
    implementation of the corresponding model from Microsoft.

    Other valid (LoraWeightInitialization) options include:
        - "random" which will use fully random initialization and is discouraged.
        - "gaussian" for Gaussian initialization.
        - "eva" for Explained Variance Adaptation (EVA) (https://arxiv.org/abs/2410.07170).
        - "loftq" for improved performance when LoRA is combined with with quantization (https://arxiv.org/abs/2310.08659).
        - "olora" for Orthonormal Low-Rank Adaptation of Large Language Models (OLoRA) (https://arxiv.org/html/2406.01775v1).
        - "pissa" for Principal Singular values and Singular vectors Adaptation (PiSSA) (https://arxiv.org/abs/2404.02948).

    For more information, see HF:
        https://github.com/huggingface/peft/blob/main/src/peft/tuners/lora/config.py
    """

    lora_task_type: TaskType = TaskType.CAUSAL_LM
    """The task type for LoRA adaptation.

    Defaults to CAUSAL_LM (Causal Language Modeling).
    """

    # Q-Lora Params
    q_lora: bool = field(default=False, metadata={"help": "Use model quantization."})
    """Whether to use quantization for LoRA (Q-LoRA).

    If True, enables quantization for more memory-efficient fine-tuning.
    """

    q_lora_bits: int = field(
        default=4, metadata={"help": "Quantization (precision) bits."}
    )
    """The number of bits to use for quantization in Q-LoRA.

    This is only used if `q_lora` is True.

    Defaults to 4-bit quantization.
    """

    # FIXME the names below use the bnb short for bits-and-bytes
    # If we consider wrapping more quantization libraries a better
    # naming convention should be applied.
    bnb_4bit_quant_type: str = field(
        default="fp4", metadata={"help": "4-bit quantization type (fp4 or nf4)."}
    )
    """The type of 4-bit quantization to use.

    Can be 'fp4' (float point 4) or 'nf4' (normal float 4).
    """

    llm_int8_skip_modules: Optional[list[str]] = field(
        default=None,
        metadata={"help": "Modules that we do not want to convert in 8-bit"},
    )
    """ An explicit list of the modules that we do not want to convert in 8-bit.

    """

    use_bnb_nested_quant: bool = field(
        default=False, metadata={"help": "Use nested quantization."}
    )
    """Whether to use nested quantization.

    Nested quantization can provide additional memory savings.
    """

    bnb_4bit_quant_storage: str = field(
        default="uint8",
        metadata={"help": "Storage type to pack the quantized 4-bit params."},
    )
    """The storage type for packing quantized 4-bit parameters.

    Defaults to 'uint8' for efficient storage.
    """

    bnb_4bit_compute_dtype: str = field(
        default="float32",
        metadata={"help": "The compute type of the quantized parameters."},
    )
    """Compute type of the quantized parameters.
    It can be different than the input type, e.g., it can be set to a lower precision
    for improved speed.

    The string will be converted to the corresponding torch.dtype.

    Valid string options are:
    - "float32" for 32-bit floating point
    - "float16" for 16-bit floating point
    - "bfloat16" for brain floating point
    - "float64" for 64-bit floating point

    Defaults to "float16" for half precision.
    """

    peft_save_mode: PeftSaveMode = PeftSaveMode.ADAPTER_ONLY
    """How to save the final model during PEFT training.

    This option is only used if `TrainingParams.save_final_model` is True.
    By default, only the model adapter is saved to reduce disk usage.
    Options are defined in the `PeftSaveMode` enum and include:
    - ADAPTER_ONLY: Only save the model adapter.
    - ADAPTER_AND_BASE_MODEL: Save the base model in addition to the adapter.
    - MERGED: Merge the adapter and base model's weights and save as a single model.
    """

    def to_lora(self) -> LoraConfig:
        """Creates a configuration for LoRA via HF's peft library."""
        if self.lora_init_weights == LoraWeightInitialization.RANDOM:
            init_lora_weights = False
        elif self.lora_init_weights == LoraWeightInitialization.DEFAULT:
            init_lora_weights = True
        else:
            init_lora_weights = self.lora_init_weights.value

        # LoraConfig's target_modules is type Optional[Union[list[str], str]], but
        # since OmegaConf doesn't support a union between a list and a primitive type,
        # our field's type is Optional[list[str]].
        #
        # This is special handling for the "all-linear" special case.
        # See: https://huggingface.co/docs/peft/en/package_reference/lora#peft.LoraConfig.target_modules
        target_modules = self.lora_target_modules
        if target_modules == ["all-linear"]:
            target_modules = "all-linear"

        return LoraConfig(
            r=self.lora_r,
            lora_alpha=self.lora_alpha,
            lora_dropout=self.lora_dropout,
            target_modules=target_modules,
            target_parameters=self.lora_target_parameters,
            modules_to_save=self.lora_modules_to_save,
            bias=self.lora_bias,  # type: ignore
            task_type=self.lora_task_type,
            init_lora_weights=init_lora_weights,
        )

    def to_bits_and_bytes(self) -> BitsAndBytesConfig:
        """Creates a configuration for quantized models via BitsAndBytes.

        The resulting configuration uses the instantiated peft parameters.
        """
        quantization_config = BitsAndBytesConfig(
            load_in_4bit=self.q_lora_bits == 4,
            load_in_8bit=self.q_lora_bits == 8,
            llm_int8_skip_modules=self.llm_int8_skip_modules,
            bnb_4bit_compute_dtype=self.bnb_4bit_compute_dtype,
            bnb_4bit_quant_type=self.bnb_4bit_quant_type,
            bnb_4bit_use_double_quant=self.use_bnb_nested_quant,
            bnb_4bit_quant_storage=self.bnb_4bit_quant_storage,
        )
        return quantization_config



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random
from pprint import pformat
from typing import Any, Callable, Optional, Union

import lm_eval.loggers.utils as lm_harness_log_utils
import numpy as np
import torch
from lm_eval import evaluate as lm_harness_evaluate
from lm_eval.api.group import ConfigurableGroup
from lm_eval.api.registry import get_model as lm_harness_get_model_class
from lm_eval.api.task import Task
from lm_eval.loggers import WandbLogger
from lm_eval.tasks import get_task_dict as lm_harness_get_task_dict

from oumi.builders import build_processor, build_tokenizer
from oumi.builders.models import is_image_text_llm_using_model_name
from oumi.core.configs import (
    EvaluationConfig,
    GenerationParams,
    InferenceEngineType,
    LMHarnessTaskParams,
    ModelParams,
    RemoteParams,
)
from oumi.core.distributed import is_world_process_zero
from oumi.core.evaluation.evaluation_result import EvaluationResult
from oumi.utils.logging import logger

# Used to set the few-shot seed for lm_eval.api.task.Task. The value is consistent with
# LM Harness `simple_evaluate`'s default `fewshot_random_seed` = 1234.
FEW_SHOT_SEED = 1234

########################################################################################
# How to map LM Harness `model_args` to Oumi's `ModelParams` and `GenerationParams`?   #
# Which LM Harness `model` types (hf, vllm, etc) support each parameter?               #
# ------------------- | -------------- | -- | ---- | ------------- | -------- | ------ #
# LM Harness          | Oumi           | LM Harness `model`                            #
# `model_args`        | `model_params` | hf | vllm | hf-multimodal | vllm-vlm | remote #
# ------------------- | -------------- | -- | ---- | ------------- | -------- | ------ #
# trust_remote_code   |                | Υ  | Υ    | Υ             | Υ        | Y      #
# pretrained          | model_name     | Υ  | Υ    | Υ             | Υ        | Y      #
# dtype               | torch_dtype    | Υ  | Υ    | Υ             | Υ        | Y      #
# max_length          |model_max_length| Υ  | Υ    | Υ             | Υ        | Y      #
# tokenizer           | tokenizer_name | Υ  | Υ    | Υ             | Υ        | Y      #
# peft                | adapter_model  | Υ  |      | Υ             |          |        #
# parallelize         | shard_for_eval | Υ  |      | Υ             |          |        #
# device_map          |                | ?? |      | ??            |          |        #
# attn_implementation |                | ?? |      | ??            |          |        #
# ------------------- | -------------- | -- | ---- | ------------- | -------- | ------ #
# max_images          |                | NA | NA   | Υ             | Υ        | NA     #
# interleave          |                | NA | NA   | Υ             | Υ        | NA     #
# convert_img_format  |                | NA | NA   | Υ             |          | NA     #
# image_token_id      |                | NA | NA   | Υ             |          | NA     #
# image_string        |                | NA | NA   | Υ             |          | NA     #
########################################################################################
# LM Harness          | Oumi `generat- | LM Harness `model`                            #
# `model_args`        | ion_params`    | hf | vllm | hf-multimodal | vllm-vlm | remote #
# ------------------- | -------------- | -- | ---- | ------------- | -------- | ------ #
# batch_size          |                | Υ  | Υ    | Υ             | Υ        | Y      #
########################################################################################

########################################################################################
# How to map LM Harness `model_args` (specifically the ones related to a remote        #
# inference engine) to Oumi's `remote_params`?                                         #
# ----------------------- | ---------------------------------------------------------- #
# LM Harness `model_args` | Oumi `remote_params`                                       #
# ----------------------- | ---------------------------------------------------------- #
# base_url                |  api_url                                                   #
# num_concurrent          |  num_workers                                               #
# max_retries             |  max_retries                                               #
# timeout                 |  connection_timeout                                        #
########################################################################################

########################################################################################
# Mapping of LM Harness `model` types to the corresponding class and file              #
# --------------------------|---------------------|----------------------------------- #
# LM Harness `model`        | Class               | File in lm-evaluation-harness repo #
# (= inference engine)      | name                | located under lm_eval/models/...   #
# --------------------------|---------------------|----------------------------------- #
# hf                        | HFLM                | huggingface.py                     #
# vllm                      | VLLM                | vllm_causallms.py                  #
# hf-multimodal             | HFMultimodalLM      | hf_vlms.py                         #
# vllm-vlm                  | VLLM_VLM            | vllm_vlms.py                       #
# local-completions         | LocalCompletionsAPI | openai_completions.py              #
########################################################################################


def _generate_lm_harness_model_args(
    lm_harness_model: str,
    is_multimodal: bool,
    device: str,
    model_params: ModelParams,
    generation_params: GenerationParams,
    inference_engine_type: InferenceEngineType,
    inference_remote_params: Optional[RemoteParams],
) -> dict[str, Any]:
    """Converts Oumi's ModelParams to LM Harness model arguments."""
    # List of all LM Harness model arguments:
    # https://github.com/EleutherAI/lm-evaluation-harness/blob/365fcda9b85bbb6e0572d91976b8daf409164500/lm_eval/models/huggingface.py#L66

    # If batch size isn't specified, we set it to "auto", which will let LM Harness
    # automatically select the largest batch size that will fit in memory.
    # https://github.com/EleutherAI/lm-evaluation-harness/blob/main/docs/interface.md
    batch_size = generation_params.batch_size or "auto"

    # Arguments used across all engines and modalities.
    model_args_dict = {
        "trust_remote_code": model_params.trust_remote_code,
        "pretrained": model_params.model_name,
        "dtype": model_params.torch_dtype or model_params.torch_dtype_str,
        "max_length": model_params.model_max_length,
        "batch_size": batch_size,
        "max_batch_size": None,
        "device": device,
    }
    if model_params.tokenizer_name:
        model_args_dict["tokenizer"] = model_params.tokenizer_name

    # Add NATIVE inference engine's additional parameters.
    if inference_engine_type == InferenceEngineType.NATIVE:
        if model_args_dict["batch_size"] == "auto":
            # NATIVE (hf) inference engine can NOT accept auto batch size.
            model_args_dict["batch_size"] = 1
        model_args_dict["parallelize"] = model_params.shard_for_eval
        model_args_dict["device_map"] = model_params.device_map
        if model_params.adapter_model:
            model_args_dict["peft"] = model_params.adapter_model
        if model_params.attn_implementation:
            model_args_dict["attn_implementation"] = model_params.attn_implementation

    # Add REMOTE inference engine's additional parameters.
    if inference_engine_type == InferenceEngineType.REMOTE:
        if not inference_remote_params:
            raise ValueError(
                "The `REMOTE` inference engine requires `inference_remote_params`."
            )
        model_args_dict["base_url"] = inference_remote_params.api_url
        if inference_remote_params.num_workers > 0:
            model_args_dict["num_concurrent"] = inference_remote_params.num_workers
        if inference_remote_params.max_retries > 0:
            model_args_dict["max_retries"] = inference_remote_params.max_retries
        if inference_remote_params.connection_timeout > 0:
            model_args_dict["timeout"] = int(inference_remote_params.connection_timeout)

    # Add multi-modal related parameters.
    # details at https://github.com/EleutherAI/lm-evaluation-harness/releases/tag/v0.4.5
    if is_multimodal:
        # FIXME OPE-355 To remove `max_images=1` limit
        model_args_dict |= {"max_images": 1, "interleave": True}

        # Only applicable to hf-multimodal (NOT vllm-vlm).
        if lm_harness_model == "hf-multimodal":
            model_args_dict["convert_img_format"] = True

            tokenizer = build_tokenizer(model_params)
            processor = build_processor(
                model_params.model_name,
                tokenizer,
                trust_remote_code=model_params.trust_remote_code,
                processor_kwargs=model_params.processor_kwargs,
            )
            if image_token := processor.image_token:
                model_args_dict["image_string"] = image_token
            if image_token_id := processor.image_token_id:
                model_args_dict["image_token_id"] = image_token_id

    # Handle extra model_kwargs (construction arguments).
    # Towards OPE-564.
    if model_params.model_kwargs:
        for key in ["load_in_4bit", "load_in_8bit", "max_memory_per_gpu"]:
            if key in model_params.model_kwargs:
                model_args_dict[key] = model_params.model_kwargs[key]
        # TODO: load_in_8bit, load_in_4bit are deprecated and will be removed in
        # future versions of HF. Integrate via PeftConfig.
    return model_args_dict


def _apply_to_all_tasks(
    task_dict: dict[Union[str, ConfigurableGroup], Union[Task, dict]],
    fn: Callable,
    fn_kwargs: Optional[dict[str, Any]] = None,
) -> None:
    """Apply the provided function `fn` to all tasks in the `task_dict`."""
    fn_kwargs = fn_kwargs or {}
    for task_obj in task_dict.values():
        if isinstance(task_obj, dict):
            _apply_to_all_tasks(task_obj, fn, fn_kwargs)
        elif isinstance(task_obj, Task):
            fn(task_obj, **fn_kwargs)
        else:
            raise ValueError(f"Expected `lm_eval.api.task.Task` but got: {task_obj}")


def _get_task_dict(
    task_params: LMHarnessTaskParams,
) -> dict[Union[str, ConfigurableGroup], Union[Task, dict]]:
    """Get a dictionary of LM Harness tasks, given Oumi's `task_params`."""
    if not task_params.task_name:
        raise ValueError("The `task_name` must be specified for LM Harness evaluation.")
    task_dict: dict = lm_harness_get_task_dict(task_params.task_name)

    # Sanity checks for `task_dict`.
    if not task_dict:
        raise ValueError(f"Task `{task_params.task_name}` not available in LM Harness.")
    elif len(task_dict) > 1:
        raise ValueError(
            "Unexpected `task_dict` from LM Harness, consisting of multiple tasks, "
            f"while a single task ({task_params.task_name}) was requested: {task_dict}."
        )
    else:
        assert len(task_dict) == 1
        task_name = next(iter(task_dict))
        if isinstance(task_name, ConfigurableGroup):
            task_name: str = task_name.group_name
        if task_name != task_params.task_name:
            raise ValueError(
                f"Inconsistent task naming. Task `{task_params.task_name}` was "
                f"requested, but LM Harness returned the `task_dict`: {task_dict}."
            )

    # Apply the following function to each task in the task_dict, in order to overwrite
    # the default parameters with the ones specified in `task_params`.
    def overwrite_task_params(task: Task) -> None:
        # Set the number of few shots to be added to the prompt.
        task.set_config(key="num_fewshot", value=task_params.num_fewshot)

        # Set the random seed (for reproducibility and consistency with LM Harness).
        task.set_fewshot_seed(seed=FEW_SHOT_SEED)

    _apply_to_all_tasks(task_dict, fn=overwrite_task_params)

    return task_dict


def _set_random_seeds(random_seed, numpy_random_seed, torch_random_seed) -> None:
    """Setting random seeds for reproducibility and consistency with LM Harness."""
    if random_seed is not None:
        random.seed(random_seed)

    if numpy_random_seed is not None:
        np.random.seed(numpy_random_seed)

    if torch_random_seed is not None:
        torch.manual_seed(torch_random_seed)


def evaluate(
    task_params: LMHarnessTaskParams,
    config: EvaluationConfig,
    random_seed: Optional[int] = 0,
    numpy_random_seed: Optional[int] = 1234,
    torch_random_seed: Optional[int] = 1234,
) -> EvaluationResult:
    """Evaluates a model using the LM Evaluation Harness framework (EleutherAI).

    For detailed documentation, we refer you to the following readme:
    https://github.com/EleutherAI/lm-evaluation-harness

    Args:
        task_params: The LM Harness parameters to use for evaluation.
        config: The evaluation configuration.
        random_seed: The random seed to use for python's `random` package.
        numpy_random_seed: The numpy random seed to use for reproducibility.
        torch_random_seed: The torch random seed to use for reproducibility.

    Note for random seeds (random_seed, numpy_random_seed, torch_random_seed):
        These have been set to be consistent with LM Harness' simple_evaluate().
        See: lm-evaluation-harness/blob/main/lm_eval/evaluator.py

    Returns:
        The evaluation results (dict of metric names and their corresponding values).
    """
    _set_random_seeds(
        random_seed=random_seed,
        numpy_random_seed=numpy_random_seed,
        torch_random_seed=torch_random_seed,
    )

    if torch.cuda.is_available():
        # CUDA device may be overwritten if `accelerate launch`,
        # or `parallelize=True` are used.
        device = "cuda:0"
    elif torch.backends.mps.is_available():
        device = "mps"
    else:
        device = "cpu"
        logger.warning("No GPU available.")

    # Identify whether the model is multi-modal.
    is_multimodal = is_image_text_llm_using_model_name(
        model_name=config.model.model_name,
        trust_remote_code=config.model.trust_remote_code,
    )

    # Identify the proper LM Harness model (`lm_harness_model`) to use.
    if config.inference_engine == InferenceEngineType.NATIVE:
        lm_harness_model = "hf-multimodal" if is_multimodal else "hf"
        if device.startswith("cuda"):
            logger.warning(
                "Since you have GPU support, it is highly recommended that you set "
                "the `inference_engine` to `VLLM`, instead of the `NATIVE`, for faster "
                "evaluation."
            )
    elif config.inference_engine == InferenceEngineType.VLLM:
        lm_harness_model = "vllm-vlm" if is_multimodal else "vllm"
        if not device.startswith("cuda"):
            raise ValueError("The `VLLM` inference_engine requires a CUDA-enabled GPU.")
    elif config.inference_engine == InferenceEngineType.REMOTE:
        lm_harness_model = "local-completions"
    else:
        raise ValueError(
            f"Unsupported inference engine type: {config.inference_engine}. "
            "Our integration with the `lm_harness` evaluation backend supports "
            "the `NATIVE`, `VLLM` and `REMOTE` inference_engine types."
        )

    # Instantiate an LM Harness task dictionary.
    task_dict = _get_task_dict(task_params)
    logger.info(f"\tLM Harness `task_params`:\n{pformat(task_params)}")
    logger.info(f"\tLM Harness `task_dict`:\n{pformat(task_dict)}")

    # Instantiate an LM Harness language model (lm).
    lm_harness_model_params = _generate_lm_harness_model_args(
        lm_harness_model=lm_harness_model,
        is_multimodal=is_multimodal,
        device=device,
        model_params=config.model,
        generation_params=config.generation,
        inference_engine_type=config.inference_engine,
        inference_remote_params=config.inference_remote_params,
    )
    logger.info(f"\tLM Harness `model_params`:\n{pformat(lm_harness_model_params)}")
    lm_class = lm_harness_get_model_class(lm_harness_model)
    lm = lm_class(**lm_harness_model_params)

    logger.info("Starting evaluation...")

    lm_eval_output = lm_harness_evaluate(
        lm,
        task_dict,
        log_samples=task_params.log_samples or False,
        limit=task_params.num_samples,
        apply_chat_template=is_multimodal,
        **task_params.eval_kwargs,  # type: ignore
    )

    # Metrics are only available on the main process, and `None` on others.
    if not is_world_process_zero():
        return EvaluationResult()

    assert lm_eval_output is not None
    task_name = task_params.task_name
    metric_dict = lm_eval_output["results"][task_name]  # type: ignore
    logger.info(f"{task_name}'s metric dict is {pformat(metric_dict)}")

    if config.enable_wandb:
        project_name = os.environ.get("WANDB_PROJECT", "oumi")
        logger.info(f"Logging to Weights and Biases project: '{project_name}'")
        wandb_logger = WandbLogger(
            init_args={
                "project": project_name,
                "name": config.run_name,
                "job_type": "eval",
            },
            config_args={"config": config},
        )
        wandb_logger.post_init(lm_eval_output)
        wandb_logger.log_eval_result()

    # The LM Harness backend's task configuration is a dictionary which
    # includes: the number of samples, the number of few-shots, task version(s),
    # the prompt(s) text, model/git hashes, seeds, and the special tokens used
    # by the tokenizer (such as `pad`, `eos`, `bos, and `eot`).
    backend_task_config = lm_eval_output

    # The LM Harness backend's results is a dictionary that includes all
    # evaluation metrics, which are oftentimes grouped (in `groups`) by a theme
    # or a classification category.
    backend_results = {
        key: backend_task_config.pop(key)
        for key in ["results", "groups"]
        if key in backend_task_config
    }

    # Add LM Harness-specific configuration settings to the results.
    backend_task_config.setdefault("config", {})

    # Add configuration settings related to the model.
    backend_task_config["config"]["model"] = lm_harness_model
    backend_task_config["config"]["model_args"] = lm_harness_model_params
    if hasattr(lm, "get_model_info"):
        backend_task_config["config"].update(lm.get_model_info())

    # Add configuration settings related to the task.
    backend_task_config["config"]["task_params"] = task_params
    backend_task_config["config"]["task_dict"] = task_dict

    # Add other configuration settings.
    backend_task_config["git_hash"] = lm_harness_log_utils.get_git_commit_hash()
    lm_harness_log_utils.add_env_info(backend_task_config)
    lm_harness_log_utils.add_tokenizer_info(backend_task_config, lm)

    return EvaluationResult(
        task_name=task_params.task_name,
        task_result=backend_results,
        backend_config=backend_task_config,
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from pathlib import Path

from huggingface_hub import HFCacheInfo
from huggingface_hub.utils import scan_cache_dir


@dataclass
class CachedItem:
    """A class representing a cached HuggingFace repository item."""

    repo_id: str
    size_bytes: int
    size: str
    repo_path: Path
    last_modified: str
    last_accessed: str
    repo_type: str
    nb_files: int


def list_hf_cache() -> list[CachedItem]:
    """List all cached items in the HuggingFace cache directory."""
    cache_dir: HFCacheInfo = scan_cache_dir()
    cached_items: list[CachedItem] = []
    for repo in cache_dir.repos:
        repo_id: str = repo.repo_id
        size_bytes: float = repo.size_on_disk
        repo_path: Path = repo.repo_path
        last_modified: str = repo.last_modified_str
        last_accessed: str = repo.last_accessed_str
        repo_type: str = repo.repo_type
        nb_files: int = repo.nb_files
        size: str = format_size(size_bytes)
        cached_items.append(
            CachedItem(
                repo_id=repo_id,
                size_bytes=size_bytes,
                size=size,
                repo_path=repo_path,
                last_modified=last_modified,
                last_accessed=last_accessed,
                repo_type=repo_type,
                nb_files=nb_files,
            )
        )
    return cached_items


def format_size(size_bytes: float) -> str:
    """Format bytes to human readable string."""
    # Simple implementation
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size_bytes < 1024:
            return f"{size_bytes:.1f}{unit}"
        size_bytes /= 1024
    return f"{size_bytes:.1f}PB"



from unittest.mock import Mock, patch

import pytest

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import JobState, JobStatus
from oumi.core.registry import REGISTRY, RegistryType
from oumi.launcher.clients.local_client import LocalClient
from oumi.launcher.clouds.local_cloud import LocalCloud
from oumi.launcher.clusters.local_cluster import LocalCluster


#
# Fixtures
#
@pytest.fixture
def mock_local_client():
    with patch("oumi.launcher.clouds.local_cloud.LocalClient") as client:
        yield client


@pytest.fixture
def mock_local_cluster():
    with patch("oumi.launcher.clouds.local_cloud.LocalCluster") as cluster:
        yield cluster


def _get_default_job(cloud: str) -> JobConfig:
    resources = JobResources(
        cloud=cloud,
        region="us-central1",
        zone=None,
        accelerators="A100-80GB",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={},
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup="pip install -r requirements.txt",
        run="./hello_world.sh",
    )


#
# Tests
#
def test_local_cloud_up_cluster(mock_local_client, mock_local_cluster):
    cloud = LocalCloud()
    mock_client = Mock(spec=LocalClient)
    mock_local_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=LocalCluster)
    mock_local_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="cluster.name",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = _get_default_job("local")
    job_status = cloud.up_cluster(job, "cluster.name")
    mock_local_client.assert_called_once()
    mock_local_cluster.assert_called_once_with("cluster.name", mock_client)
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_local_cloud_up_cluster_no_name(mock_local_client, mock_local_cluster):
    cloud = LocalCloud()
    mock_client = Mock(spec=LocalClient)
    mock_local_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=LocalCluster)
    mock_local_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="cluster.name",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = _get_default_job("local")
    job_status = cloud.up_cluster(job, None)
    mock_local_client.assert_called_once()
    mock_local_cluster.assert_called_once_with("local", mock_client)
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_local_cloud_list_clusters(mock_local_client):
    cloud = LocalCloud()
    mock_client = Mock(spec=LocalClient)
    mock_local_client.side_effect = [mock_client, mock_client]
    # Check that there are no initial clusters.
    assert [] == cloud.list_clusters()
    job = _get_default_job("local")
    cloud.up_cluster(job, None)
    cloud.up_cluster(job, "new cluster")
    cloud.up_cluster(job, "local")
    clusters = cloud.list_clusters()
    expected_clusters = [
        "local",
        "new cluster",
    ]
    cluster_names = [cluster.name() for cluster in clusters]
    cluster_names.sort()
    assert cluster_names == expected_clusters


def test_local_cloud_get_cluster_empty(mock_local_client):
    cloud = LocalCloud()
    # Check that there are no initial clusters.
    assert cloud.get_cluster("local") is None


def test_local_cloud_get_cluster_success(mock_local_client):
    cloud = LocalCloud()
    mock_client = Mock(spec=LocalClient)
    mock_local_client.side_effect = [mock_client, mock_client]
    job = _get_default_job("local")
    cloud.up_cluster(job, None)
    cloud.up_cluster(job, "new cluster")
    cloud.up_cluster(job, "local")
    expected_clusters = [
        "local",
        "new cluster",
    ]
    for name in expected_clusters:
        cluster = cloud.get_cluster(name)
        assert cluster is not None
        assert cluster.name() == name


def test_local_cloud_get_cluster_fails(mock_local_client):
    cloud = LocalCloud()
    mock_client = Mock(spec=LocalClient)
    mock_local_client.side_effect = [mock_client, mock_client]
    job = _get_default_job("local")
    cloud.up_cluster(job, None)
    cloud.up_cluster(job, "new cluster")
    cloud.up_cluster(job, "local")
    assert cloud.get_cluster("nonexistent") is None


def test_local_cloud_builder_registered():
    assert REGISTRY.contains("local", RegistryType.CLOUD)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from oumi.builders.inference_engines import build_inference_engine
from oumi.core.configs.inference_config import InferenceConfig
from oumi.core.configs.inference_engine_type import InferenceEngineType
from oumi.core.configs.params.synthesis_params import (
    GeneralSynthesisParams,
    GeneratedAttribute,
    GeneratedAttributePostprocessingParams,
    TextMessage,
)
from oumi.core.synthesis.attribute_formatter import AttributeFormatter
from oumi.core.types.conversation import Conversation, Message
from oumi.utils.logging import logger


class AttributeSynthesizer:
    """Synthesizes values for a generated attribute based on the given samples.

    Args:
        params: The parameters for the attribute synthesizer.
        inference_config: The configuration for the inference engine.
    """

    def __init__(
        self,
        params: GeneralSynthesisParams,
        inference_config: InferenceConfig,
    ):
        """Initialize the synthesizer."""
        self._params = params
        self._formatter = AttributeFormatter(params)

        self._inference_engine = build_inference_engine(
            engine_type=inference_config.engine or InferenceEngineType.NATIVE,
            model_params=inference_config.model,
            remote_params=inference_config.remote_params,
        )
        self._inference_config = inference_config

    def synthesize(
        self,
        samples: list[dict],
        generated_attribute: GeneratedAttribute,
    ) -> list[dict[str, str]]:
        """Synthesize a value for the generated attribute.

        Order will be identical to the order of the samples.

        Args:
            samples: The samples to synthesize values for.
            generated_attribute: The generated attribute to synthesize a value for.

        Returns:
            A list of dictionaries, one for each sample, with the generated attribute
            value added to the dictionary.
        """
        inference_conversations: list[Conversation] = []
        for sample in samples:
            inference_conversations.append(
                self._format_instructions(
                    sample,
                    generated_attribute.instruction_messages,
                )
            )

        inference_results = self._inference_engine.infer(
            inference_conversations,
            inference_config=self._inference_config,
        )

        original_responses = self._extract_response(inference_results)
        if not generated_attribute.postprocessing_params:
            records = [
                {generated_attribute.id: unpostprocessed_response}
                for unpostprocessed_response in original_responses
            ]
            return records

        keep_original = (
            generated_attribute.postprocessing_params.keep_original_text_attribute
        )
        if keep_original:
            records = [
                {generated_attribute.id: unpostprocessed_response}
                for unpostprocessed_response in original_responses
            ]
        else:
            records = [{} for _ in original_responses]

        for i in range(len(original_responses)):
            new_id = generated_attribute.postprocessing_params.id
            original_response = original_responses[i]
            new_response = original_response
            try:
                new_response = self._postprocess_sample(
                    original_response, generated_attribute.postprocessing_params
                )
            except ValueError as e:
                logger.warning(
                    f"Error postprocessing inference result: {e}. Leaving as-is and "
                    "skipping."
                )
            finally:
                records[i][new_id] = new_response

        return records

    def _extract_response(
        self,
        inference_conversations: list[Conversation],
    ) -> list[str]:
        """Get the inference results from the inference conversations.

        If the inference result is not a string, an empty string will be returned.
        """
        return [
            inference_result.messages[-1].content
            if isinstance(inference_result.messages[-1].content, str)
            else ""
            for inference_result in inference_conversations
        ]

    def _format_instructions(
        self,
        sample: dict,
        instruction_messages: list[TextMessage],
    ) -> Conversation:
        """Format the instructions for the sample."""
        new_messages = []
        for turn in instruction_messages:
            if not isinstance(turn.content, str):
                new_messages.append(turn)
                continue

            formatted_content = self._formatter.format(
                sample,
                turn.content,
                missing_values_allowed=False,
            )
            new_message = Message(
                role=turn.role,
                content=formatted_content,
            )
            new_messages.append(new_message)

        return Conversation(messages=new_messages)

    def _postprocess_sample(
        self,
        response: str,
        postprocessing_params: GeneratedAttributePostprocessingParams,
    ) -> str:
        """Postprocess the response, removing extraneous text.

        Order of operations:
        1. If regex is provided, use the first match.
        2. Cut off everything before the first occurrence of the prefix and after the
        last occurrence of the suffix.
        3. Strip whitespace.
        4. Add prefix and suffix to what remains.

        Args:
            response: The response to postprocess.
            postprocessing_params: The postprocessing parameters.

        Returns:
            The postprocessed response.
        """
        if postprocessing_params.regex:
            match = re.search(postprocessing_params.regex, response)
            if match:
                response = match.group(0)

        # Cut off prefix and suffix
        if postprocessing_params.cut_prefix:
            prefix_loc = response.find(postprocessing_params.cut_prefix)
            if prefix_loc != -1:
                response = response[
                    prefix_loc + len(postprocessing_params.cut_prefix) :
                ]
        if postprocessing_params.cut_suffix:
            suffix_loc = response.rfind(postprocessing_params.cut_suffix)
            if suffix_loc != -1:
                response = response[:suffix_loc]

        # Strip whitespace
        if postprocessing_params.strip_whitespace:
            response = response.strip()

        # Add prefix and suffix
        if postprocessing_params.added_prefix:
            response = postprocessing_params.added_prefix + response
        if postprocessing_params.added_suffix:
            response = response + postprocessing_params.added_suffix

        return response



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the LengthAnalyzer."""

from unittest.mock import Mock

import pytest

from oumi.core.analyze.length_analyzer import LengthAnalyzer
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.utils.analysis_utils import conversation_to_dataframes


def _single_message_conversation(text):
    return Conversation(messages=[Message(role=Role.USER, content=text)])


def _count_analysis_columns(df):
    """Count the number of analysis columns in a DataFrame."""
    analysis_suffixes = [
        "_char_count",
        "_word_count",
        "_sentence_count",
        "_token_count",
    ]
    return len(
        [
            col
            for col in df.columns
            if any(col.endswith(suffix) for suffix in analysis_suffixes)
        ]
    )


def test_char_count():
    """Test character count functionality."""
    analyzer = LengthAnalyzer(
        char_count=True, word_count=False, sentence_count=False, token_count=False
    )
    conv = _single_message_conversation("Hello, world!")
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)
    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )
    assert result_df.iloc[0]["text_content_char_count"] == 13
    # Only char_count should be present
    assert _count_analysis_columns(result_df) == 1


def test_word_count():
    """Test word count functionality."""
    analyzer = LengthAnalyzer(
        char_count=False, word_count=True, sentence_count=False, token_count=False
    )
    conv = _single_message_conversation("Hello world! This is a test.")
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)
    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )
    assert result_df.iloc[0]["text_content_word_count"] == 6
    # Only word_count should be present
    assert _count_analysis_columns(result_df) == 1


def test_sentence_count():
    """Test sentence count functionality."""
    analyzer = LengthAnalyzer(
        char_count=False, word_count=False, sentence_count=True, token_count=False
    )
    conv = _single_message_conversation("Hello world! This is a test. How are you?")
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)

    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )
    assert result_df.iloc[0]["text_content_sentence_count"] == 3
    # Only sentence_count should be present
    assert _count_analysis_columns(result_df) == 1


def test_analyzer_instantiation():
    """Test analyzer can be instantiated with different parameter combinations."""
    # Test with defaults
    analyzer = LengthAnalyzer()
    conv = _single_message_conversation("Hello, world!")
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)

    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )
    assert result_df.iloc[0]["text_content_char_count"] == 13
    assert result_df.iloc[0]["text_content_word_count"] == 2
    assert result_df.iloc[0]["text_content_sentence_count"] == 1
    assert "token_count" not in result_df.columns

    # Test with custom parameters
    analyzer = LengthAnalyzer(
        char_count=True, word_count=False, sentence_count=True, token_count=False
    )
    conv = _single_message_conversation("Hello, world!")
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)

    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )
    assert result_df.iloc[0]["text_content_char_count"] == 13
    assert "word_count" not in result_df.columns
    assert result_df.iloc[0]["text_content_sentence_count"] == 1
    assert "token_count" not in result_df.columns

    # Test with partial parameters (some defaults, some overridden)
    analyzer = LengthAnalyzer(char_count=False, word_count=True)
    conv = _single_message_conversation("Hello, world!")
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)

    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )
    assert "char_count" not in result_df.columns
    assert result_df.iloc[0]["text_content_word_count"] == 2
    assert result_df.iloc[0]["text_content_sentence_count"] == 1  # Default True
    assert "token_count" not in result_df.columns  # Default False


def test_token_count():
    """Test token count functionality."""
    # Test token count with tokenizer only (default: includes special tokens)
    mock_tokenizer = Mock()
    mock_tokenizer.apply_chat_template.return_value = "<prompt>"
    mock_tokenizer.encode.return_value = [0, 1, 2, 3, 4, 5, 2]  # 7 tokens

    analyzer = LengthAnalyzer(
        char_count=False,
        word_count=False,
        sentence_count=False,
        token_count=True,
        tokenizer=mock_tokenizer,
    )
    conv = _single_message_conversation("Hello, world!")
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)

    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )
    assert result_df.iloc[0]["text_content_token_count"] == 7
    # analyze calls tokenizer once per field
    assert mock_tokenizer.encode.call_count == 1
    # Check that it was called with the message text
    mock_tokenizer.encode.assert_any_call("Hello, world!", add_special_tokens=True)

    # Test without special tokens (explicitly set to False)
    mock_tokenizer_no_special = Mock()
    mock_tokenizer_no_special.encode.return_value = [1, 2, 3, 4, 5]  # 5 tokens

    analyzer_no_special = LengthAnalyzer(
        char_count=False,
        word_count=False,
        sentence_count=False,
        token_count=True,
        tokenizer=mock_tokenizer_no_special,
        include_special_tokens=False,
    )
    conv = _single_message_conversation("Hello, world!")
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)
    result_df = analyzer_no_special.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )
    assert result_df.iloc[0]["text_content_token_count"] == 5
    # Check that it was called without special tokens
    mock_tokenizer_no_special.encode.assert_any_call(
        "Hello, world!", add_special_tokens=False
    )

    # Test without tokenizer (should raise ValueError)
    with pytest.raises(ValueError, match="tokenizer must be provided"):
        analyzer_no_tokenizer = LengthAnalyzer(
            char_count=False,
            word_count=False,
            sentence_count=False,
            token_count=True,
            # No tokenizer
        )
        _, test_df = conversation_to_dataframes(conv, "test_conv", 0)
        analyzer_no_tokenizer.analyze_sample(
            test_df, schema={"text_content": {"content_type": "text"}}
        )

    # Test with tokenizer but token_count=False (should not call tokenizer)
    mock_tokenizer_unused = Mock()
    analyzer_unused = LengthAnalyzer(
        char_count=True,
        word_count=False,
        sentence_count=False,
        token_count=False,  # Token count disabled
        tokenizer=mock_tokenizer_unused,
    )
    conv = _single_message_conversation("Hello, world!")
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)
    result_df = analyzer_unused.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )
    # Should not call tokenizer since token_count=False
    mock_tokenizer_unused.encode.assert_not_called()
    # Should still compute char_count
    assert result_df.iloc[0]["text_content_char_count"] == 13


def test_conversation_level_token_count():
    """Test that conversation-level token count is computed correctly with tokenizer."""
    mock_tokenizer = Mock()
    mock_tokenizer.apply_chat_template.return_value = "<prompt>"
    # 6 tokens for each message; 10 tokens for conversation
    mock_tokenizer.encode.side_effect = [
        [0, 1, 2, 3, 4, 5],
        [0, 1, 2, 3, 4, 5],
        list(range(10)),
    ]

    # Create analyzer without dataset
    analyzer = LengthAnalyzer(
        char_count=False,
        word_count=False,
        sentence_count=False,
        token_count=True,
        tokenizer=mock_tokenizer,
    )

    # Create a conversation with multiple messages
    conv = Conversation(
        messages=[
            Message(role=Role.USER, content="Hello, how are you?"),
            Message(role=Role.ASSISTANT, content="I am doing well, thank you!"),
        ]
    )

    # Analyze the conversation
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)
    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )

    # Check that field-level token count is computed for each message
    assert "text_content_token_count" in result_df.columns
    # Each message should have 6 tokens
    assert result_df.iloc[0]["text_content_token_count"] == 6
    assert result_df.iloc[1]["text_content_token_count"] == 6

    # Verify that encode was used for field-level token count
    # Two message encodes (one per row)
    assert mock_tokenizer.encode.call_count == 2


def test_conversation_level_token_count_without_dataset():
    """Test that conversation-level token count is computed without a dataset using
    tokenizer chat template directly."""
    mock_tokenizer = Mock()
    mock_tokenizer.apply_chat_template.return_value = "<prompt>"
    mock_tokenizer.encode.side_effect = [
        [0, 1, 2, 3, 4, 5],
        [0, 1, 2, 3, 4, 5],
        list(range(8)),
    ]

    # Create analyzer WITHOUT dataset
    analyzer = LengthAnalyzer(
        char_count=False,
        word_count=False,
        sentence_count=False,
        token_count=True,
        tokenizer=mock_tokenizer,
        # No dataset parameter
    )

    # Create a conversation with multiple messages
    conv = Conversation(
        messages=[
            Message(role=Role.USER, content="Hello, how are you?"),
            Message(role=Role.ASSISTANT, content="I am doing well, thank you!"),
        ]
    )

    # Analyze the conversation
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)
    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )

    # Check that field-level token count is computed for each message
    assert result_df.iloc[0]["text_content_token_count"] == 6
    assert result_df.iloc[1]["text_content_token_count"] == 6
    # Two message encodes (one per row)
    assert mock_tokenizer.encode.call_count == 2


def test_conversation_level_metrics_aggregation():
    """Test that conversation-level metrics are correctly aggregated from message-level
    metrics."""
    # Test that char, word, and sentence counts are aggregated from message-level
    # results
    mock_tokenizer = Mock()
    mock_tokenizer.apply_chat_template.return_value = "<prompt>"
    mock_tokenizer.encode.side_effect = [
        [0, 1, 2, 3, 4, 5],
        [0, 1, 2, 3, 4, 5],
        list(range(10)),
    ]

    # Create analyzer with all metrics enabled
    analyzer = LengthAnalyzer(
        char_count=True,
        word_count=True,
        sentence_count=True,
        token_count=True,
        tokenizer=mock_tokenizer,
    )

    # Create a conversation with multiple messages
    conv = Conversation(
        messages=[
            Message(
                role=Role.USER, content="Hello, how are you?"
            ),  # 18 chars, 4 words, 1 sentence
            Message(
                role=Role.ASSISTANT, content="I am doing well, thank you!"
            ),  # 26 chars, 6 words, 1 sentence
        ]
    )

    # Analyze the conversation
    _, test_df = conversation_to_dataframes(conv, "test_conv", 0)
    result_df = analyzer.analyze_sample(
        test_df, schema={"text_content": {"content_type": "text"}}
    )

    # Check field-level metrics for each message
    # First message: "Hello, how are you?" - 19 chars, 4 words, 1 sentence
    assert result_df.iloc[0]["text_content_char_count"] == 19
    assert result_df.iloc[0]["text_content_word_count"] == 4
    assert result_df.iloc[0]["text_content_sentence_count"] == 1
    assert result_df.iloc[0]["text_content_token_count"] == 6

    # Second message: "I am doing well, thank you!" - 27 chars, 6 words, 1 sentence
    assert result_df.iloc[1]["text_content_char_count"] == 27
    assert result_df.iloc[1]["text_content_word_count"] == 6
    assert result_df.iloc[1]["text_content_sentence_count"] == 1
    assert result_df.iloc[1]["text_content_token_count"] == 6

    # Two message encodes (one per row)
    assert mock_tokenizer.encode.call_count == 2



import tempfile

from oumi.core.configs import (
    DataParams,
    DatasetParams,
    DatasetSplitParams,
    ModelParams,
    TuningConfig,
    TuningParams,
)
from oumi.tune import tune


def test_tune():
    """Test tuning multiple hyperparameters simultaneously."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        config = TuningConfig(
            data=DataParams(
                train=DatasetSplitParams(
                    datasets=[
                        DatasetParams(
                            dataset_name="debug_sft",
                            dataset_kwargs={"dataset_size": 10},
                        )
                    ]
                ),
                validation=DatasetSplitParams(
                    datasets=[
                        DatasetParams(
                            dataset_name="debug_sft", dataset_kwargs={"dataset_size": 5}
                        )
                    ]
                ),
            ),
            model=ModelParams(
                model_name="openai-community/gpt2",
                model_max_length=1024,
                trust_remote_code=True,
                tokenizer_pad_token="<|endoftext|>",
            ),
            tuning=TuningParams(
                tunable_training_params={
                    "learning_rate": {
                        "type": "loguniform",
                        "low": 1e-5,
                        "high": 1e-4,
                    },
                    "per_device_train_batch_size": {
                        "type": "categorical",
                        "choices": [2, 4],
                    },
                },
                fixed_training_params={
                    "max_steps": 2,
                    "enable_wandb": False,
                    "enable_tensorboard": False,
                    "enable_mlflow": False,
                },
                tunable_peft_params={
                    "lora_r": {
                        "type": "categorical",
                        "choices": [4, 8],
                    },
                    "lora_alpha": {
                        "type": "categorical",
                        "choices": [16, 32],
                    },
                },
                output_dir=output_temp_dir,
                n_trials=2,
            ),
        )

        tune(config)


def test_tune_multi_objective():
    """Test multi-objective hyperparameter optimization."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        config = TuningConfig(
            data=DataParams(
                train=DatasetSplitParams(
                    datasets=[
                        DatasetParams(
                            dataset_name="debug_sft",
                            dataset_kwargs={"dataset_size": 10},
                        )
                    ]
                ),
                validation=DatasetSplitParams(
                    datasets=[
                        DatasetParams(
                            dataset_name="debug_sft", dataset_kwargs={"dataset_size": 5}
                        )
                    ]
                ),
            ),
            model=ModelParams(
                model_name="openai-community/gpt2",
                model_max_length=1024,
                trust_remote_code=True,
                tokenizer_pad_token="<|endoftext|>",
            ),
            tuning=TuningParams(
                tunable_training_params={
                    "learning_rate": {
                        "type": "categorical",
                        "choices": [1e-5, 5e-5],
                    },
                },
                fixed_training_params={
                    "max_steps": 2,
                    "enable_wandb": False,
                    "enable_tensorboard": False,
                    "enable_mlflow": False,
                },
                output_dir=output_temp_dir,
                evaluation_metrics=["eval_loss", "eval_mean_token_accuracy"],
                evaluation_direction=["minimize", "maximize"],
                n_trials=2,
            ),
        )

        tune(config)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TypeVar

from oumi.core.configs import JobConfig
from oumi.core.launcher import BaseCloud, BaseCluster, JobStatus
from oumi.core.registry import register_cloud_builder
from oumi.launcher.clients.sky_client import SkyClient
from oumi.launcher.clusters.sky_cluster import SkyCluster

T = TypeVar("T")


class SkyCloud(BaseCloud):
    """A resource pool capable of creating clusters using Sky Pilot."""

    @property
    def _client(self) -> SkyClient:
        """Returns the SkyClient instance."""
        # Instantiating a SkyClient imports sky.
        # Delay sky import: https://github.com/oumi-ai/oumi/issues/1605
        if not self._sky_client:
            self._sky_client = SkyClient()
        return self._sky_client

    def __init__(self, cloud_name: str):
        """Initializes a new instance of the SkyCloud class."""
        self._cloud_name = cloud_name
        self._sky_client: Optional[SkyClient] = None

    def _get_clusters_by_class(self, cloud_class: type[T]) -> list[BaseCluster]:
        """Gets the appropriate clusters of type T."""
        # Delay sky import: https://github.com/oumi-ai/oumi/issues/1605
        import sky

        return [
            SkyCluster(cluster["name"], self._client)
            for cluster in self._client.status()
            if (
                isinstance(cluster["handle"].launched_resources.cloud, cloud_class)
                and cluster["status"] == sky.ClusterStatus.UP
            )
        ]

    def up_cluster(self, job: JobConfig, name: Optional[str], **kwargs) -> JobStatus:
        """Creates a cluster and starts the provided Job."""
        job_status = self._client.launch(job, name, **kwargs)
        cluster = self.get_cluster(job_status.cluster)
        if not cluster:
            raise RuntimeError(f"Cluster {job_status.cluster} not found.")
        return cluster.get_job(job_status.id)

    def get_cluster(self, name) -> Optional[BaseCluster]:
        """Gets the cluster with the specified name, or None if not found."""
        clusters = self.list_clusters()
        for cluster in clusters:
            if cluster.name() == name:
                return cluster
        return None

    def list_clusters(self) -> list[BaseCluster]:
        """Lists the active clusters on this cloud."""
        # Delay sky import: https://github.com/oumi-ai/oumi/issues/1605
        import sky

        if self._cloud_name == SkyClient.SupportedClouds.GCP.value:
            return self._get_clusters_by_class(sky.clouds.GCP)
        elif self._cloud_name == SkyClient.SupportedClouds.RUNPOD.value:
            return self._get_clusters_by_class(sky.clouds.RunPod)
        elif self._cloud_name == SkyClient.SupportedClouds.LAMBDA.value:
            return self._get_clusters_by_class(sky.clouds.Lambda)
        elif self._cloud_name == SkyClient.SupportedClouds.AWS.value:
            return self._get_clusters_by_class(sky.clouds.AWS)
        elif self._cloud_name == SkyClient.SupportedClouds.AZURE.value:
            return self._get_clusters_by_class(sky.clouds.Azure)
        raise ValueError(f"Unsupported cloud: {self._cloud_name}")


@register_cloud_builder("runpod")
def runpod_cloud_builder() -> SkyCloud:
    """Builds a SkyCloud instance for runpod."""
    return SkyCloud(SkyClient.SupportedClouds.RUNPOD.value)


@register_cloud_builder("gcp")
def gcp_cloud_builder() -> SkyCloud:
    """Builds a SkyCloud instance for Google Cloud Platform."""
    return SkyCloud(SkyClient.SupportedClouds.GCP.value)


@register_cloud_builder("lambda")
def lambda_cloud_builder() -> SkyCloud:
    """Builds a SkyCloud instance for Lambda."""
    return SkyCloud(SkyClient.SupportedClouds.LAMBDA.value)


@register_cloud_builder("aws")
def aws_cloud_builder() -> SkyCloud:
    """Builds a SkyCloud instance for AWS."""
    return SkyCloud(SkyClient.SupportedClouds.AWS.value)


@register_cloud_builder("azure")
def azure_cloud_builder() -> SkyCloud:
    """Builds a SkyCloud instance for Azure."""
    return SkyCloud(SkyClient.SupportedClouds.AZURE.value)



import logging
import pathlib
import tempfile
from enum import Enum
from unittest.mock import Mock, call, patch

import pytest
import typer
from typer.testing import CliRunner

from oumi.cli.alias import AliasType
from oumi.cli.cli_utils import CONTEXT_ALLOW_EXTRA_ARGS
from oumi.cli.launch import cancel, down, logs, status, stop, up, which
from oumi.cli.launch import run as launcher_run
from oumi.core.configs import (
    DataParams,
    DatasetParams,
    DatasetSplitParams,
    ModelParams,
    TrainerType,
    TrainingConfig,
    TrainingParams,
)
from oumi.core.launcher import JobState, JobStatus
from oumi.launcher import JobConfig, JobResources
from oumi.utils.logging import logger


@pytest.fixture
def mock_fetch():
    with patch("oumi.cli.cli_utils.resolve_and_fetch_config") as m_fetch:
        m_fetch.side_effect = lambda x: x
        yield m_fetch


@pytest.fixture
def mock_alias():
    with patch("oumi.cli.launch.try_get_config_name_for_alias") as try_alias:
        yield try_alias


runner = CliRunner()


class MockPool:
    def __init__(self):
        self.mock_result = Mock()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def apply_async(self, fn, kwds):
        fn(**kwds)
        return self.mock_result


#
# Fixtures
#
@pytest.fixture
def app():
    launch_app = typer.Typer()
    launch_app.command()(down)
    launch_app.command(name="run", context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(
        launcher_run
    )
    launch_app.command()(status)
    launch_app.command()(stop)
    launch_app.command()(cancel)
    launch_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(up)
    launch_app.command()(which)
    launch_app.command()(logs)
    yield launch_app


@pytest.fixture
def mock_launcher():
    with patch("oumi.launcher", autospec=True) as launcher_mock:
        yield launcher_mock


@pytest.fixture(autouse=True)
def mock_sky_client(mock_launcher):
    mock_launcher.clients.sky_client.SkyClient.SupportedClouds = []


@pytest.fixture()
def mock_sky_tail():
    with patch("sky.tail_logs") as sky_mock:
        yield sky_mock


@pytest.fixture
def mock_pool():
    with patch("oumi.cli.launch.Pool") as pool_mock:
        mock_pool = MockPool()
        pool_mock.return_value = mock_pool
        yield pool_mock


@pytest.fixture
def mock_confirm():
    with patch("typer.confirm") as confirm_mock:
        yield confirm_mock


@pytest.fixture
def mock_version():
    with patch("oumi.utils.version_utils.version") as version_mock:
        version_mock.return_value = ""
        yield version_mock


@pytest.fixture
def mock_git_root():
    with patch("oumi.cli.launch.get_git_root_dir") as root_mock:
        root_mock.return_value = _oumi_root()
        yield root_mock


def _oumi_root() -> str:
    return "fake/oumi/root"


def _create_training_config() -> TrainingConfig:
    return TrainingConfig(
        data=DataParams(
            train=DatasetSplitParams(
                datasets=[
                    DatasetParams(
                        dataset_name="debug_sft",
                    )
                ],
            ),
        ),
        model=ModelParams(
            model_name="MlpEncoder",
            model_max_length=1024,
            trust_remote_code=True,
            tokenizer_name="gpt2",
        ),
        training=TrainingParams(
            trainer_type=TrainerType.TRL_SFT,
            max_steps=3,
            logging_steps=3,
            log_model_summary=True,
            enable_wandb=False,
            enable_tensorboard=False,
            enable_mlflow=False,
            try_resume_from_last_checkpoint=True,
            save_final_model=True,
        ),
    )


def _create_job_config(training_config_path: str) -> JobConfig:
    return JobConfig(
        name="foo",
        user="bar",
        working_dir=".",
        resources=JobResources(
            cloud="aws",
            region="us-west-1",
            zone=None,
            accelerators="A100-80GB",
            cpus="4",
            memory="64",
            instance_type=None,
            use_spot=True,
            disk_size=512,
            disk_tier="low",
        ),
        run=f"oumi launch up --config {training_config_path}",
    )


def test_launch_up_job(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        mock_log_stream = Mock()
        mock_log_stream.readline.side_effect = ["line1\n", "line2\n", ""]
        mock_cluster.get_logs_stream.return_value = mock_log_stream
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                job_yaml_path,
                "--log-level",
                "DEBUG",
            ],
        )
        mock_fetch.assert_called_once_with(job_yaml_path)
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        assert logger.level == logging.DEBUG


def test_launch_up_job_with_alias(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch, mock_alias
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        mock_alias.return_value = job_yaml_path
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        # Mock the log stream to return proper string values
        mock_log_stream = Mock()
        mock_log_stream.readline.side_effect = ["line1\n", "line2\n", ""]
        mock_cluster.get_logs_stream.return_value = mock_log_stream
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                "some_alias",
                "--log-level",
                "DEBUG",
            ],
        )
        mock_fetch.assert_called_once_with(job_yaml_path)
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        mock_alias.assert_called_once_with("some_alias", AliasType.JOB)
        assert logger.level == logging.DEBUG


def test_launch_up_job_dev_confirm(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_git_root, mock_fetch
):
    mock_version.return_value = "0.1.0.dev0"
    mock_confirm.return_value = True
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        mock_log_stream = Mock()
        mock_log_stream.readline.side_effect = ["line1\n", "line2\n", ""]
        mock_cluster.get_logs_stream.return_value = mock_log_stream
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                job_yaml_path,
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        job_config.working_dir = _oumi_root()
        mock_launcher.up.assert_called_once_with(job_config, None)


def test_launch_up_job_dev_no_confirm(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    mock_version.return_value = "0.1.0.dev0"
    mock_confirm.return_value = False
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        mock_log_stream = Mock()
        mock_log_stream.readline.side_effect = ["line1\n", "line2\n", ""]
        mock_cluster.get_logs_stream.return_value = mock_log_stream
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                job_yaml_path,
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        mock_launcher.up.assert_called_once_with(job_config, None)


def test_launch_up_job_dev_no_confirm_same_path(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    working_dir = "/foo/dir/"
    mock_version.return_value = working_dir
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.working_dir = working_dir
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        mock_log_stream = Mock()
        mock_log_stream.readline.side_effect = ["line1\n", "line2\n", ""]
        mock_cluster.get_logs_stream.return_value = mock_log_stream
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                job_yaml_path,
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        mock_launcher.up.assert_called_once_with(job_config, None)


def test_launch_up_job_no_working_dir(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.working_dir = None
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        mock_log_stream = Mock()
        mock_log_stream.readline.side_effect = ["line1\n", "line2\n", ""]
        mock_cluster.get_logs_stream.return_value = mock_log_stream
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                job_yaml_path,
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        mock_launcher.up.assert_called_once_with(job_config, None)


def test_launch_up_job_existing_cluster(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        mock_cluster.name.return_value = "cluster_id"
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.run.return_value = job_status
        mock_cluster.get_job.return_value = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        mock_log_stream = Mock()
        mock_log_stream.readline.side_effect = ["line1\n", "line2\n", ""]
        mock_cluster.get_logs_stream.return_value = mock_log_stream
        mock_cloud = Mock()
        mock_launcher.get_cloud.return_value = mock_cloud
        mock_cloud.get_cluster.return_value = mock_cluster
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                job_yaml_path,
                "--cluster",
                "cluster_id",
            ],
        )
        mock_launcher.run.assert_called_once_with(job_config, "cluster_id")
        mock_cluster.get_job.assert_has_calls([call("job_id")])


def test_launch_up_job_detach(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                job_yaml_path,
                "--detach",
            ],
        )
        mock_cluster.get_job.assert_not_called()


def test_launch_up_job_detached_local(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.resources.cloud = "local"
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="local",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_logs_stream.side_effect = NotImplementedError()
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="local",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                job_yaml_path,
                "--detach",
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])


def test_launch_up_job_sky_logs(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    class SupportedClouds(Enum):
        """Enum representing the supported clouds."""

        LOCAL = "local"

    mock_launcher.clients.sky_client.SkyClient.SupportedClouds = [SupportedClouds.LOCAL]
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.resources.cloud = "local"
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="local",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="local",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        mock_log_stream = Mock()
        mock_log_stream.readline.side_effect = ["line1\n", "line2\n", ""]
        mock_cluster.get_logs_stream.return_value = mock_log_stream
        _ = runner.invoke(
            app,
            [
                "up",
                "--config",
                job_yaml_path,
            ],
        )
        mock_cluster.get_logs_stream.assert_called_once()


def test_launch_up_job_not_found(
    app, mock_launcher, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_launcher.up.return_value = (mock_cluster, job_status)
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        with pytest.raises(FileNotFoundError) as exception_info:
            res = runner.invoke(
                app,
                [
                    "up",
                    "--config",
                    str(pathlib.Path(output_temp_dir) / "fake_path.yaml"),
                ],
            )
            if res.exception:
                raise res.exception
        assert "No such file or directory" in str(exception_info.value)


def test_launch_run_job(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_cloud = Mock()
        mock_launcher.run.return_value = job_status
        mock_launcher.get_cloud.side_effect = [mock_cloud, mock_cloud]
        mock_cloud.get_cluster.side_effect = [mock_cluster, mock_cluster]
        mock_cluster.get_logs_stream.side_effect = NotImplementedError()
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        _ = runner.invoke(
            app,
            [
                "run",
                "--config",
                job_yaml_path,
                "--cluster",
                "cluster_id",
                "-log",
                "CRITICAL",
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        mock_launcher.run.assert_called_once_with(job_config, "cluster_id")
        mock_launcher.get_cloud.assert_has_calls([call("aws")])
        mock_cloud.get_cluster.assert_has_calls([call("cluster_id")])
        mock_fetch.assert_called_once_with(job_yaml_path)
        assert logger.level == logging.CRITICAL


def test_launch_run_job_with_alias(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch, mock_alias
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        mock_alias.return_value = job_yaml_path
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_cloud = Mock()
        mock_launcher.run.return_value = job_status
        mock_launcher.get_cloud.side_effect = [mock_cloud, mock_cloud]
        mock_cloud.get_cluster.side_effect = [mock_cluster, mock_cluster]
        mock_cluster.get_logs_stream.side_effect = NotImplementedError()
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        _ = runner.invoke(
            app,
            [
                "run",
                "--config",
                "some_alias",
                "--cluster",
                "cluster_id",
                "-log",
                "CRITICAL",
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        mock_launcher.run.assert_called_once_with(job_config, "cluster_id")
        mock_launcher.get_cloud.assert_has_calls([call("aws")])
        mock_cloud.get_cluster.assert_has_calls([call("cluster_id")])
        mock_fetch.assert_called_once_with(job_yaml_path)
        mock_alias.assert_called_once_with("some_alias", AliasType.JOB)
        assert logger.level == logging.CRITICAL


def test_launch_run_job_dev_confirm(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_git_root, mock_fetch
):
    mock_version.return_value = "0.1.0.dev0"
    mock_confirm.return_value = True
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_cloud = Mock()
        mock_launcher.run.return_value = job_status
        mock_launcher.get_cloud.side_effect = [mock_cloud, mock_cloud]
        mock_cloud.get_cluster.side_effect = [mock_cluster, mock_cluster]
        mock_cluster.get_logs_stream.side_effect = NotImplementedError()
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        _ = runner.invoke(
            app,
            [
                "run",
                "--config",
                job_yaml_path,
                "--cluster",
                "cluster_id",
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        job_config.working_dir = _oumi_root()
        mock_launcher.run.assert_called_once_with(job_config, "cluster_id")
        mock_launcher.get_cloud.assert_has_calls([call("aws")])
        mock_cloud.get_cluster.assert_has_calls([call("cluster_id")])


def test_launch_run_job_dev_no_confirm(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    mock_version.return_value = "0.1.0.dev0"
    mock_confirm.return_value = False
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_cloud = Mock()
        mock_launcher.run.return_value = job_status
        mock_launcher.get_cloud.side_effect = [mock_cloud, mock_cloud]
        mock_cloud.get_cluster.side_effect = [mock_cluster, mock_cluster]
        mock_cluster.get_logs_stream.side_effect = NotImplementedError()
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        _ = runner.invoke(
            app,
            [
                "run",
                "--config",
                job_yaml_path,
                "--cluster",
                "cluster_id",
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        mock_launcher.run.assert_called_once_with(job_config, "cluster_id")
        mock_launcher.get_cloud.assert_has_calls([call("aws")])
        mock_cloud.get_cluster.assert_has_calls([call("cluster_id")])


def test_launch_run_job_detached(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_cloud = Mock()
        mock_launcher.run.return_value = job_status
        mock_launcher.get_cloud.return_value = mock_cloud
        mock_cloud.get_cluster.return_value = mock_cluster
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        _ = runner.invoke(
            app,
            [
                "run",
                "--config",
                job_yaml_path,
                "--cluster",
                "cluster_id",
                "--detach",
            ],
        )
        mock_cluster.get_job.assert_not_called()
        mock_launcher.run.assert_called_once_with(job_config, "cluster_id")
        mock_launcher.get_cloud.assert_not_called()
        mock_cloud.get_cluster.assert_not_called()


def test_launch_run_job_detached_local(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.resources.cloud = "local"
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="local",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_cloud = Mock()
        mock_launcher.run.return_value = job_status
        mock_launcher.get_cloud.side_effect = [mock_cloud, mock_cloud]
        mock_cloud.get_cluster.return_value = mock_cluster
        mock_cluster.get_logs_stream.side_effect = NotImplementedError()
        mock_cluster.get_job.return_value = JobStatus(
            id="job_id",
            cluster="local",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        _ = runner.invoke(
            app,
            [
                "run",
                "--config",
                job_yaml_path,
                "--cluster",
                "local",
            ],
        )
        mock_cluster.get_job.assert_has_calls([call("job_id")])
        mock_launcher.run.assert_called_once_with(job_config, "local")
        mock_cloud.get_cluster.assert_has_calls([call("local")])
        mock_launcher.get_cloud.assert_has_calls([call("local")])


def test_launch_run_job_no_cluster(
    app, mock_launcher, mock_pool, mock_version, mock_confirm, mock_fetch
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        train_yaml_path = str(pathlib.Path(output_temp_dir) / "train.yaml")
        config: TrainingConfig = _create_training_config()
        config.to_yaml(train_yaml_path)
        job_yaml_path = str(pathlib.Path(output_temp_dir) / "job.yaml")
        job_config = _create_job_config(train_yaml_path)
        job_config.to_yaml(job_yaml_path)
        mock_launcher.JobConfig = JobConfig
        mock_cluster = Mock()
        job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="running",
            metadata="",
            done=False,
            state=JobState.PENDING,
        )
        mock_cloud = Mock()
        mock_launcher.run.return_value = job_status
        mock_launcher.get_cloud.return_value = mock_cloud
        mock_cloud.get_cluster.return_value = mock_cluster
        mock_cluster.get_job.return_value = job_status = JobStatus(
            id="job_id",
            cluster="cluster_id",
            name="job_name",
            status="done",
            metadata="",
            done=True,
            state=JobState.SUCCEEDED,
        )
        with pytest.raises(
            ValueError, match="No cluster specified for the `run` action."
        ):
            res = runner.invoke(
                app,
                [
                    "run",
                    "--config",
                    job_yaml_path,
                ],
            )
            if res.exception:
                raise res.exception


def test_launch_down_with_cloud(app, mock_launcher, mock_pool):
    mock_cloud = Mock()
    mock_cluster = Mock()
    mock_launcher.get_cloud.return_value = mock_cloud
    mock_cloud.get_cluster.return_value = mock_cluster
    _ = runner.invoke(
        app,
        [
            "down",
            "--cluster",
            "cluster_id",
            "--cloud",
            "aws",
            "--log-level",
            "INFO",
        ],
    )
    mock_launcher.get_cloud.assert_called_once_with("aws")
    mock_cloud.get_cluster.assert_called_once_with("cluster_id")
    mock_cluster.down.assert_called_once()
    assert logger.level == logging.INFO


def test_launch_down_no_cloud(app, mock_launcher, mock_pool):
    mock_cloud1 = Mock()
    mock_cluster1 = Mock()
    mock_cloud2 = Mock()
    mock_launcher.get_cloud.side_effect = [mock_cloud1, mock_cloud2]
    mock_cloud1.get_cluster.return_value = mock_cluster1
    mock_cloud2.get_cluster.return_value = None
    mock_launcher.which_clouds.return_value = ["aws", "foo"]
    _ = runner.invoke(
        app,
        [
            "down",
            "--cluster",
            "cluster_id",
        ],
    )
    mock_launcher.get_cloud.assert_has_calls([call("aws"), call("foo")])
    mock_cloud1.get_cluster.assert_called_once_with("cluster_id")
    mock_cloud2.get_cluster.assert_called_once_with("cluster_id")
    mock_cluster1.down.assert_called_once()


def test_launch_down_multiple_clusters(app, mock_launcher, mock_pool):
    mock_cloud1 = Mock()
    mock_cluster1 = Mock()
    mock_cloud2 = Mock()
    mock_cluster2 = Mock()
    mock_launcher.get_cloud.side_effect = [mock_cloud1, mock_cloud2]
    mock_cloud1.get_cluster.return_value = mock_cluster1
    mock_cloud2.get_cluster.return_value = mock_cluster2
    mock_launcher.which_clouds.return_value = ["aws", "foo"]
    _ = runner.invoke(
        app,
        [
            "down",
            "--cluster",
            "cluster_id",
        ],
    )
    mock_launcher.get_cloud.assert_has_calls([call("aws"), call("foo")])
    mock_cloud1.get_cluster.assert_called_once_with("cluster_id")
    mock_cloud2.get_cluster.assert_called_once_with("cluster_id")
    mock_cluster1.down.assert_not_called()
    mock_cluster2.down.assert_not_called()


def test_launch_down_no_clusters(app, mock_launcher, mock_pool):
    mock_cloud1 = Mock()
    mock_cloud2 = Mock()
    mock_launcher.get_cloud.side_effect = [mock_cloud1, mock_cloud2]
    mock_cloud1.get_cluster.return_value = None
    mock_cloud2.get_cluster.return_value = None
    mock_launcher.which_clouds.return_value = ["aws", "foo"]
    _ = runner.invoke(
        app,
        [
            "down",
            "--cluster",
            "cluster_id",
        ],
    )
    mock_launcher.get_cloud.assert_has_calls([call("aws"), call("foo")])
    mock_cloud1.get_cluster.assert_called_once_with("cluster_id")
    mock_cloud2.get_cluster.assert_called_once_with("cluster_id")


def test_launch_stop_with_cloud(app, mock_launcher, mock_pool):
    mock_cloud = Mock()
    mock_cluster = Mock()
    mock_launcher.get_cloud.return_value = mock_cloud
    mock_cloud.get_cluster.return_value = mock_cluster
    _ = runner.invoke(
        app,
        [
            "stop",
            "--cluster",
            "cluster_id",
            "--cloud",
            "aws",
            "--log-level",
            "ERROR",
        ],
    )
    mock_launcher.get_cloud.assert_called_once_with("aws")
    mock_cloud.get_cluster.assert_called_once_with("cluster_id")
    mock_cluster.stop.assert_called_once()
    assert logger.level == logging.ERROR


def test_launch_stop_no_cloud(app, mock_launcher, mock_pool):
    mock_cloud1 = Mock()
    mock_cluster1 = Mock()
    mock_cloud2 = Mock()
    mock_launcher.get_cloud.side_effect = [mock_cloud1, mock_cloud2]
    mock_cloud1.get_cluster.return_value = mock_cluster1
    mock_cloud2.get_cluster.return_value = None
    mock_launcher.which_clouds.return_value = ["aws", "foo"]
    _ = runner.invoke(
        app,
        [
            "stop",
            "--cluster",
            "cluster_id",
        ],
    )
    mock_launcher.get_cloud.assert_has_calls([call("aws"), call("foo")])
    mock_cloud1.get_cluster.assert_called_once_with("cluster_id")
    mock_cloud2.get_cluster.assert_called_once_with("cluster_id")
    mock_cluster1.stop.assert_called_once()


def test_launch_stop_multiple_clusters(app, mock_launcher, mock_pool):
    mock_cloud1 = Mock()
    mock_cluster1 = Mock()
    mock_cloud2 = Mock()
    mock_cluster2 = Mock()
    mock_launcher.get_cloud.side_effect = [mock_cloud1, mock_cloud2]
    mock_cloud1.get_cluster.return_value = mock_cluster1
    mock_cloud2.get_cluster.return_value = mock_cluster2
    mock_launcher.which_clouds.return_value = ["aws", "foo"]
    _ = runner.invoke(
        app,
        [
            "stop",
            "--cluster",
            "cluster_id",
        ],
    )
    mock_launcher.get_cloud.assert_has_calls([call("aws"), call("foo")])
    mock_cloud1.get_cluster.assert_called_once_with("cluster_id")
    mock_cloud2.get_cluster.assert_called_once_with("cluster_id")
    mock_cluster1.stop.assert_not_called()
    mock_cluster2.stop.assert_not_called()


def test_launch_stop_no_clusters(app, mock_launcher, mock_pool):
    mock_cloud1 = Mock()
    mock_cloud2 = Mock()
    mock_launcher.get_cloud.side_effect = [mock_cloud1, mock_cloud2]
    mock_cloud1.get_cluster.return_value = None
    mock_cloud2.get_cluster.return_value = None
    mock_launcher.which_clouds.return_value = ["aws", "foo"]
    _ = runner.invoke(
        app,
        [
            "stop",
            "--cluster",
            "cluster_id",
        ],
    )
    mock_launcher.get_cloud.assert_has_calls([call("aws"), call("foo")])
    mock_cloud1.get_cluster.assert_called_once_with("cluster_id")
    mock_cloud2.get_cluster.assert_called_once_with("cluster_id")


def test_launch_cancel_success(app, mock_launcher, mock_pool):
    _ = runner.invoke(
        app,
        [
            "cancel",
            "--cloud",
            "cloud",
            "--cluster",
            "cluster",
            "--id",
            "job",
            "--log-level",
            "DEBUG",
        ],
    )
    mock_launcher.cancel.assert_called_once()
    assert logger.level == logging.DEBUG


def test_launch_which_success(app, mock_launcher, mock_pool):
    _ = runner.invoke(
        app,
        [
            "which",
            "-log",
            "INFO",
        ],
    )
    mock_launcher.which_clouds.assert_called_once()
    assert logger.level == logging.INFO


def test_launch_status_success(app, mock_launcher, mock_pool):
    _ = runner.invoke(
        app,
        [
            "status",
            "--cloud",
            "cloud",
            "--cluster",
            "cluster",
            "--id",
            "job",
            "-log",
            "DEBUG",
        ],
    )
    mock_launcher.status.assert_has_calls(
        [call(cloud="cloud", cluster="cluster", id="job")]
    )
    assert logger.level == logging.DEBUG


def test_launch_status_cluster_no_jobs(app, mock_launcher, mock_pool):
    mock_cluster = Mock()
    mock_cluster.name.return_value = "cluster_id"
    mock_cluster.get_jobs.return_value = []
    mock_cloud = Mock()
    mock_launcher.get_cloud.return_value = mock_cloud
    mock_launcher.status.return_value = {"cloud_id": []}
    mock_cloud.list_clusters.return_value = [mock_cluster]
    result = runner.invoke(
        app,
        [
            "status",
        ],
    )
    mock_launcher.status.assert_has_calls([call(cloud=None, cluster=None, id=None)])
    assert "Cloud: cloud_id" in result.stdout
    assert "Cluster: cluster_id" in result.stdout
    assert "No matching jobs found." in result.stdout


# Tests for logs command
def test_launch_logs_with_cloud_success(app, mock_launcher):
    """Test logs command with specific cloud - success case."""
    mock_cloud = Mock()
    mock_cluster = Mock()
    mock_log_stream = Mock()
    mock_launcher.get_cloud.return_value = mock_cloud
    mock_cloud.get_cluster.return_value = mock_cluster
    mock_cluster.get_logs_stream.return_value = mock_log_stream

    with patch("oumi.cli.launch._tail_logs") as mock_tail_logs:
        _ = runner.invoke(
            app,
            [
                "logs",
                "--cluster",
                "test_cluster",
                "--cloud",
                "aws",
                "--job-id",
                "job_123",
            ],
        )

    mock_launcher.get_cloud.assert_called_once_with("aws")
    mock_cloud.get_cluster.assert_called_once_with("test_cluster")
    mock_cluster.get_logs_stream.assert_called_once_with("test_cluster", "job_123")
    mock_tail_logs.assert_called_once_with(mock_log_stream, None)


def test_launch_logs_with_cloud_cluster_not_found(app, mock_launcher, mock_pool):
    """Test logs command with specific cloud - cluster not found."""
    mock_cloud = Mock()
    mock_launcher.get_cloud.return_value = mock_cloud
    mock_cloud.get_cluster.return_value = None

    result = runner.invoke(
        app,
        [
            "logs",
            "--cluster",
            "test_cluster",
            "--cloud",
            "aws",
        ],
    )

    assert result.exit_code != 0
    assert "Cluster test_cluster not found" in result.stdout


def test_launch_logs_no_cloud_single_match(app, mock_launcher):
    """Test logs command without cloud - single cluster found."""
    mock_cloud1 = Mock()
    mock_cluster1 = Mock()
    mock_cloud2 = Mock()
    mock_log_stream = Mock()
    mock_launcher.which_clouds.return_value = ["aws", "gcp"]
    mock_launcher.get_cloud.side_effect = [mock_cloud1, mock_cloud2]
    mock_cloud1.get_cluster.return_value = mock_cluster1
    mock_cloud2.get_cluster.return_value = None
    mock_cluster1.get_logs_stream.return_value = mock_log_stream

    with patch("oumi.cli.launch._tail_logs") as mock_tail_logs:
        _ = runner.invoke(
            app,
            [
                "logs",
                "--cluster",
                "test_cluster",
            ],
        )

    mock_launcher.which_clouds.assert_called_once()
    mock_launcher.get_cloud.assert_has_calls([call("aws"), call("gcp")])
    mock_cloud1.get_cluster.assert_called_once_with("test_cluster")
    mock_cloud2.get_cluster.assert_called_once_with("test_cluster")
    mock_cluster1.get_logs_stream.assert_called_once_with("test_cluster", None)
    mock_tail_logs.assert_called_once_with(mock_log_stream, None)


def test_launch_logs_no_cloud_multiple_matches(app, mock_launcher, mock_pool):
    """Test logs command without cloud - multiple clusters found."""
    mock_cloud1 = Mock()
    mock_cluster1 = Mock()
    mock_cloud2 = Mock()
    mock_cluster2 = Mock()
    mock_launcher.which_clouds.return_value = ["aws", "gcp"]
    mock_launcher.get_cloud.side_effect = [mock_cloud1, mock_cloud2]
    mock_cloud1.get_cluster.return_value = mock_cluster1
    mock_cloud2.get_cluster.return_value = mock_cluster2

    result = runner.invoke(
        app,
        [
            "logs",
            "--cluster",
            "test_cluster",
        ],
    )

    assert result.exit_code != 0
    assert "Multiple clusters found with name test_cluster" in result.stdout


def test_launch_logs_no_cloud_no_matches(app, mock_launcher):
    """Test logs command without cloud - no clusters found."""
    mock_cloud1 = Mock()
    mock_cloud2 = Mock()
    mock_launcher.which_clouds.return_value = ["aws", "gcp"]
    mock_launcher.get_cloud.side_effect = [mock_cloud1, mock_cloud2]
    mock_cloud1.get_cluster.return_value = None
    mock_cloud2.get_cluster.return_value = None

    result = runner.invoke(
        app,
        [
            "logs",
            "--cluster",
            "test_cluster",
        ],
    )

    assert result.exit_code != 0
    assert "Cluster test_cluster not found" in result.stdout


def test_launch_logs_with_output_file(app, mock_launcher):
    """Test logs command with output file specified."""
    mock_cloud = Mock()
    mock_cluster = Mock()
    mock_log_stream = Mock()
    mock_launcher.get_cloud.return_value = mock_cloud
    mock_cloud.get_cluster.return_value = mock_cluster
    mock_cluster.get_logs_stream.return_value = mock_log_stream

    with patch("oumi.cli.launch._tail_logs") as mock_tail_logs:
        _ = runner.invoke(
            app,
            [
                "logs",
                "--cluster",
                "test_cluster",
                "--cloud",
                "aws",
                "--output-filepath",
                "/tmp/logs.txt",
            ],
        )

    mock_cluster.get_logs_stream.assert_called_once_with("test_cluster", None)
    mock_tail_logs.assert_called_once_with(mock_log_stream, "/tmp/logs.txt")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

from oumi.core.configs import TrainingParams
from oumi.core.registry import REGISTRY


def build_reward_functions(config: TrainingParams) -> list[Callable]:
    """Builds the reward functions."""
    result: list[Callable] = []
    if config.reward_functions is not None:
        # Import to ensure GRPO reward functions are added to REGISTRY.
        import oumi.datasets.grpo.rewards as grpo_rewards  # noqa: F401

        function_names = [name for name in config.reward_functions if name]
        for name in function_names:
            reward_function = REGISTRY.get_reward_function(name)
            if not reward_function:
                raise KeyError(
                    f"reward_function `{name}` was not found in the registry."
                )
            result.append(reward_function)

    return result



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A callback to detect NaN/INF metric values."""

import copy
from typing import Optional, Union

import numpy as np
import transformers

from oumi.core.callbacks.base_trainer_callback import BaseTrainerCallback
from oumi.core.configs import TrainingParams
from oumi.utils.logging import logger

_LOGS_KWARG = "logs"


class NanInfDetectionCallback(BaseTrainerCallback):
    """Trainer callback to detect abnormal values (NaN, INF) of selected metrics.

    For example, `NaN` loss value is an almost certain indication of a training process
    going badly, in which cases it's best to detect the condition early, and fail.
    """

    def __init__(
        self,
        metrics: list[str],
    ):
        """Initializes the NanInfDetectionCallback.

        Args:
            metrics: The list of metrics to monitor.
        """
        self._metrics = copy.deepcopy(metrics)

    def on_log(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called after logging the last logs."""
        metrics_dict = kwargs.pop(_LOGS_KWARG, None)
        if not metrics_dict:
            return

        # Now check for NaN or Inf.
        for metric in self._metrics:
            metric_val = metrics_dict.get(metric, None)
            if metric_val is not None and (
                np.isnan(metric_val) or np.isinf(metric_val)
            ):
                error_message = (
                    "NaN" if np.isnan(metric_val) else "INF"
                ) + f" is detected for the '{metric}' metric."
                logger.error(error_message)
                raise RuntimeError(error_message)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
from pathlib import Path
from typing import Optional


def get_git_revision_hash() -> Optional[str]:
    """Get the current git revision hash.

    Returns:
        Optional[str]: The current git revision hash, or None if it cannot be retrieved.
    """
    try:
        return subprocess.check_output(
            ["git", "rev-parse", "HEAD"],
            cwd=Path(__file__).parent,
            stderr=subprocess.DEVNULL,
            text=True,
        ).strip()
    except (subprocess.CalledProcessError, FileNotFoundError):
        return None


def get_git_tag() -> Optional[str]:
    """Get the current git tag.

    Returns:
        Optional[str]: The current git tag, or None if it cannot be retrieved.
    """
    try:
        return subprocess.check_output(
            ["git", "describe", "--tags", "--abbrev=0", "--exact-match"],
            cwd=Path(__file__).parent,
            stderr=subprocess.DEVNULL,
            text=True,
        ).strip()
    except (subprocess.CalledProcessError, FileNotFoundError):
        return None


def get_git_root_dir() -> Optional[Path]:
    """Get the root directory of the current git repository.

    Returns:
        Optional[str]: The root directory of the current git repository, or None if it
        cannot be retrieved.
    """
    try:
        dir_path = Path(
            subprocess.check_output(
                ["git", "rev-parse", "--show-toplevel"],
                stderr=subprocess.DEVNULL,
                text=True,
            ).strip()
        )
        return dir_path if dir_path.is_dir() else None
    except (subprocess.CalledProcessError, FileNotFoundError):
        return None
